@@ -1,17 +1,19 @@
-// Copyright (c) 2015  David Muse
+// Copyright (c) 1999-2018 David Muse
 // See the file COPYING for more information
 
 #if defined(USE_STRING)
 	#include <string>
 #elif defined(USE_STRINGBUFFER)
 	#include <rudiments/stringbuffer.h>
+#elif defined(USE_BYTEBUFFER)
+	#include <rudiments/bytebuffer.h>
 #endif
-#include <rudiments/datetime.h>
 #include <rudiments/stdio.h>
+#include "../test/test.cpp"
 
 #define ITEMS 1024*1024*10
-//#define ITERS 160
-#define ITERS 1
+//#define ITERS 10
+#define ITERS 160
 
 int main(int argc, char **argv) {
 
@@ -19,13 +21,17 @@ int main(int argc, char **argv) {
 	std::string	str;
 #elif defined(USE_STRINGBUFFER)
 	stringbuffer	str;
+	stdoutput.printf("initial size: %d\n",str.getInitialSize());
+#elif defined(USE_BYTEBUFFER)
+	bytebuffer	str;
+	stdoutput.printf("initial size: %d\n",str.getInitialSize());
 #endif
 	datetime	start;
-	datetime	finish;
+	datetime	end;
 
 	char	block[1024];
 	for (uint64_t b=0; b<sizeof(block); b++) {
-		block[b]='a';
+		block[b]='a'+(b%10);
 	}
 
 	for (uint64_t i=1; i<=1024; i=i*2) {
@@ -40,14 +46,10 @@ int main(int argc, char **argv) {
 			for (uint64_t k=0; k<ITEMS/i; k++) {
 				str.append(block,i);
 			}
-
 			str.clear();
-
-			stdoutput.printf(".");
 		}
-		finish.getSystemDateAndTime();
+		end.getSystemDateAndTime();
 
-		stdoutput.printf("\n%lld seconds\n\n",
-				finish.getEpoch()-start.getEpoch());
+		displayTime(&start,&end);
 	}
 }
