@@ -1,4 +1,4 @@
-// Copyright (c) 2015  David Muse
+// Copyright (c) 1999-2018 David Muse
 // See the file COPYING for more information
 
 #if defined(USE_MAP)
@@ -6,10 +6,10 @@
 #elif defined(USE_DICTIONARY)
 	#include <rudiments/dictionary.h>
 #endif
-#include <rudiments/datetime.h>
 #include <rudiments/stdio.h>
+#include "../test/test.cpp"
 
-#define ITEMS 1024*4
+#define ITEMS 1024*100
 #define ITERS 80
 
 int main(int argc, char **argv) {
@@ -18,9 +18,10 @@ int main(int argc, char **argv) {
 	std::map<uint64_t,uint64_t>	d;
 #elif defined(USE_DICTIONARY)
 	dictionary<uint64_t,uint64_t>	d;
+	d.setTrackInsertionOrder(false);
 #endif
 	datetime	start;
-	datetime	finish;
+	datetime	end;
 
 	stdoutput.printf("appending %d items (%d times)\n",ITEMS,ITERS);
 
@@ -39,8 +40,7 @@ int main(int argc, char **argv) {
 
 		stdoutput.printf(".");
 	}
-	finish.getSystemDateAndTime();
+	end.getSystemDateAndTime();
 
-	stdoutput.printf("\n%lld seconds\n\n",
-			finish.getEpoch()-start.getEpoch());
+	displayTime(&start,&end);
 }
