@@ -1,14 +1,18 @@
+// Copyright (c) 1999-2018 David Muse
+// See the COPYING file for more information.
+
 #include <rudiments/codetree.h>
 #include <rudiments/file.h>
 #include <rudiments/charstring.h>
 #include <rudiments/character.h>
 #include <rudiments/linkedlist.h>
+#include <rudiments/dictionary.h>
 #include <rudiments/stdio.h>
 
 
 #if (defined(_MSC_VER) && (_MSC_VER <= 1300)) || \
 	(defined(__USLC__) && !defined(__GNUC__)) || \
-	defined(__VMS)
+	defined(__VMS) || defined(__DECCXX)
 
 	// degenerate debug macros for really incapable compilers
 	// (old microsoft compilers, unixware native compiler, openvms compiler)
@@ -42,6 +46,7 @@ static const char	LETTER='l';
 static const char	LOWERCASELETTER='w';
 static const char	UPPERCASELETTER='u';
 static const char	DIGIT='i';
+static const char	NONPRINTABLECHARACTER='p';
 static const char	SET='s';
 static const char	NONTERMINAL='n';
 static const char	BREAK='b';
@@ -61,6 +66,7 @@ static const char	*END="e";
 static const char	*VALUE="v";
 static const char	*CASE="c";
 static const char	*RECURSIVE="r";
+static const char	*ALIAS="a";
 
 static const char	INLINE='i';
 static const char	LITERAL='l';
@@ -71,12 +77,15 @@ static const char	NONE='n';
 static const char	YES='y';
 
 static const char	STX=0x2;
+static const char	EOT=0x4;
 
 class codetreegrammarprivate {
 	friend class codetreegrammar;
 	private:
 		char	_currentattribute;
 		bool	_hasrecursivebreak;
+
+		dictionary<const char *, domnode *>	_definitions;
 };
 
 codetreegrammar::codetreegrammar() : xmldom(false) {
@@ -92,7 +101,7 @@ codetreegrammar::~codetreegrammar() {
 bool codetreegrammar::tagStart(const char *ns, const char *name) {
 
 	// get the first letter of the name
-	char	ch=name[0];
+	char	ch=*name;
 
 	// the first letters of these collide with other tags and require
 	// special handling...
@@ -102,6 +111,8 @@ bool codetreegrammar::tagStart(const char *ns, const char *name) {
 		ch=LOWERCASELETTER;
 	} else if (!charstring::compare(name,"digit")) {
 		ch=DIGIT;
+	} else if (!charstring::compare(name,"nonprintablecharacter")) {
+		ch=NONPRINTABLECHARACTER;
 	}
 
 	// replace the name
@@ -111,17 +122,30 @@ bool codetreegrammar::tagStart(const char *ns, const char *name) {
 	return xmldom::tagStart(ns,newname);
 }
 
+bool codetreegrammar::tagEnd(const char *ns, const char *name) {
+
+	bool	retval=xmldom::tagEnd(ns,name);
+
+	if (*name==GRAMMAR[0]) {
+		buildDefinitionDictionary();
+		buildNonTerminalNodeAssociations(
+			getRootNode()->getFirstTagChild(GRAMMAR));
+	}
+
+	return retval;
+}
+
 bool codetreegrammar::attributeName(const char *name) {
 
 	// the first letters of this one collides with other names and requires
 	// special handling...
 	// tag collides with type
-	char	ch=name[0];
+	char	ch=*name;
 	if (!charstring::compare(name,"tag")) {
 		ch=TAG[0];
 	}
 
-	pvt->_currentattribute=name[0];
+	pvt->_currentattribute=*name;
 
 	if (ch=='r') {
 		pvt->_hasrecursivebreak=true;
@@ -144,7 +168,7 @@ bool codetreegrammar::attributeValue(const char *value) {
 	// the first letters of this one collides with other values and requires
 	// special handling...
 	// line collides with literal
-	char	ch=value[0];
+	char	ch=*value;
 	if (!charstring::compare(value,"line")) {
 		ch=LINE;
 	}
@@ -160,6 +184,52 @@ bool codetreegrammar::hasRecursiveBreak() {
 	return pvt->_hasrecursivebreak;
 }
 
+void codetreegrammar::buildDefinitionDictionary() {
+
+	// Build a dictionary, mapping the names of nonterminals to the
+	// associated <definition> tag.  This makes lookups much faster
+	// in buildNonTerminalNodeAssociations, but more importantly, in
+	// codetree::write().
+
+	pvt->_definitions.clear();
+	for (domnode *node=getRootNode()->
+				getFirstTagChild(GRAMMAR)->
+				getFirstTagChild(DEFINITION);
+			!node->isNullNode();
+			node=node->getNextTagSibling(DEFINITION)) {
+		pvt->_definitions.setValue(node->getAttributeValue(NAME),node);
+	}
+}
+
+void codetreegrammar::buildNonTerminalNodeAssociations(domnode *node) {
+
+	// For each <nonterminal> tag, this method attaches a pointer to the
+	// corresponding <definition> tag.  That way, when the parser encounters
+	// a nonterminal and needs to look up the definition it can do it
+	// directly rather than having to search the tree.  This yields a
+	// significant performance improvement at the cost of only a single
+	// pass through the tree to assign the pointers.
+
+	// if it's a nonterminal node
+	if (*(node->getName())==NONTERMINAL) {
+
+		// find the associated definition and attach it to this node
+		node->setPrivateData((void *)pvt->_definitions.
+				getValue(node->getAttributeValue(NAME)));
+	}
+
+	// process children
+	for (domnode *child=node->getFirstTagChild();
+		!child->isNullNode(); child=child->getNextTagSibling()) {
+		buildNonTerminalNodeAssociations(child);
+	}
+}
+
+domnode *codetreegrammar::getDefinition(const char *name) {
+	domnode	*def=pvt->_definitions.getValue(name);
+	return (def)?def:getNullNode();
+}
+
 struct break_t {
 	const char	*value;
 	size_t		valuelength;
@@ -171,7 +241,7 @@ class codetreeprivate {
 	friend class codetree;
 	private:
 		codetreegrammar		_grammar;
-		xmldomnode		*_grammartag;
+		domnode			*_grammartag;
 		bool			_error;
 		uint32_t		_depth;
 		const char		*_indentstring;
@@ -182,7 +252,7 @@ class codetreeprivate {
 		const char		*_finalcodeposition;
 		uint8_t			_debuglevel;
 		stringbuffer		_excbuffer;
-		xmldomnode		*_excnode;
+		domnode		*_excnode;
 		bool			_endofstring;
 
 		bool					_break;
@@ -218,7 +288,7 @@ void codetree::setDebugLevel(uint8_t debuglevel) {
 bool codetree::parse(const char *input,
 			const char *grammar,
 			const char *startsymbol,
-			xmldomnode *output,
+			domnode *output,
 			const char **codeposition) {
 
 	// load the input grammar
@@ -233,7 +303,7 @@ bool codetree::parse(const char *input,
 bool codetree::parse(const char *input,
 			codetreegrammar *grammar,
 			const char *startsymbol,
-			xmldomnode *output,
+			domnode *output,
 			const char **codeposition) {
 
 	// verify the grammar
@@ -245,17 +315,15 @@ bool codetree::parse(const char *input,
 	// get the namespace
 	pvt->_ns=pvt->_grammartag->getAttributeValue(NAMESPACE);
 
-	// build non-terminal node associations
-	// (see method for explanation)
+	// (re)set the start symbol
 	pvt->_grammartag->setPrivateData((void *)
-		pvt->_grammartag->getFirstTagChild(DEFINITION,NAME,
-							startsymbol));
-	buildNonTerminalNodeAssociations(pvt->_grammartag);
+		((codetreegrammar *)(pvt->_grammartag->getTree()))->
+						getDefinition(startsymbol));
 
 	// initialize a node for processing exceptions
-	pvt->_excnode=new xmldomnode(output->getTree(),
+	pvt->_excnode=new domnode(output->getTree(),
 					output->getNullNode(),
-					TAG_XMLDOMNODETYPE,
+					TAG_DOMNODETYPE,
 					NULL,"excnode",NULL);
 
 	// re-init the error
@@ -284,49 +352,24 @@ bool codetree::parse(const char *input,
 	return retval;
 }
 
-void codetree::buildNonTerminalNodeAssociations(xmldomnode *node) {
-
-	// For each <nonterminal> tag, this method attaches a pointer to the
-	// corresponding <definition> tag.  That way, when the parser encounters
-	// a nonterminal and needs to look up the definition it can do it
-	// directly rather than having to search the tree.  This yields a
-	// significant performance improvement at the cost of only a single
-	// pass through the tree to assign the pointers.
-
-	// if it's a nonterminal node
-	if (node->getName()[0]=='n') {
-
-		// find the associated definition and attach it to this node
-		node->setPrivateData((void *)
-			pvt->_grammartag->getFirstTagChild(DEFINITION,NAME,
-						node->getAttributeValue(NAME)));
-	}
-
-	// process children
-	for (xmldomnode *child=node->getFirstTagChild();
-		!child->isNullNode(); child=child->getNextTagSibling()) {
-		buildNonTerminalNodeAssociations(child);
-	}
-}
-
-char codetree::getSymbolType(xmldomnode *nt) {
+char codetree::getSymbolType(domnode *nt) {
 	const char	*symboltype=nt->getAttributeValue(TYPE);
 	if (charstring::isNullOrEmpty(symboltype)) {
 		return INLINE;
 	}
-	return symboltype[0];
+	return *symboltype;
 }
 
-bool codetree::isTag(xmldomnode *nt) {
+bool codetree::isTag(domnode *nt) {
 	const char	*tag=nt->getAttributeValue(TAG);
 	if (charstring::isNullOrEmpty(tag)) {
 		return false;
 	}
-	return (tag[0]==YES);
+	return (*tag==YES);
 }
 
-bool codetree::parseChild(xmldomnode *grammarnode,
-				xmldomnode *treeparent,
+bool codetree::parseChild(domnode *grammarnode,
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 
@@ -356,7 +399,7 @@ bool codetree::parseChild(xmldomnode *grammarnode,
 	bool		retval=true;
 	const char	*name=grammarnode->getName();
 	if (name) {
-		switch (name[0]) {
+		switch (*name) {
 			case CONCATENATION:
 				if (!parseConcatenation(grammarnode,
 							treeparent,
@@ -429,6 +472,15 @@ bool codetree::parseChild(xmldomnode *grammarnode,
 					retval=false;
 				}
 				break;
+			case NONPRINTABLECHARACTER:
+				if (!parseNonPrintableCharacter(
+							grammarnode,
+							treeparent,
+							codeposition,
+							localntbuffer)) {
+					retval=false;
+				}
+				break;
 			case SET:
 				if (!parseSet(grammarnode,
 							treeparent,
@@ -470,9 +522,9 @@ bool codetree::parseChild(xmldomnode *grammarnode,
 	if (retval) {
 
 		// if the next node is an exception...
-		xmldomnode	*sibling=grammarnode->getNextTagSibling();
+		domnode	*sibling=grammarnode->getNextTagSibling();
 		name=sibling->getName();
-		if (!sibling->isNullNode() && name && name[0]=='e') {
+		if (!sibling->isNullNode() && name && *name=='e') {
 
 			// create some local buffers
 			pvt->_excbuffer.clear();
@@ -521,40 +573,19 @@ bool codetree::parseChild(xmldomnode *grammarnode,
 	return retval;
 }
 
-bool codetree::parseConcatenation(xmldomnode *grammarnode,
-				xmldomnode *treeparent,
+bool codetree::parseConcatenation(domnode *grammarnode,
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 	debugPrintIndent(4);
 	debugPrintf(4,"concatenation... {\n");
 
 	// all children must parse successfully
-	for (xmldomnode *child=grammarnode->getFirstTagChild();
+	for (domnode *child=grammarnode->getFirstTagChild();
 		!child->isNullNode(); child=child->getNextTagSibling()) {
 
-		// if we hit the end of the string, then that's only ok if
-		// the rest of the children are option, repetition, break
-		// or exception
-		if (pvt->_endofstring) {
-
-			for (xmldomnode *rchild=child->getNextTagSibling();
-				!rchild->isNullNode();
-				rchild=rchild->getNextTagSibling()) {
-
-				const char	*name=rchild->getName();
-				if (!name || (name[0]!=OPTION &&
-						name[0]!=REPETITION &&
-						name[0]!=BREAK &&
-						name[0]!=EXCEPTION)) {
-					debugPrintIndent(4);
-					debugPrintf(4,
-						"} concatenation failed\n");
-					return false;
-				}
-			}
-		}
-
-		if (!parseChild(child,treeparent,codeposition,ntbuffer)) {
+		if ((pvt->_endofstring && !endOfStringOk(child)) || 
+			!parseChild(child,treeparent,codeposition,ntbuffer)) {
 			debugPrintIndent(4);
 			debugPrintf(4,"} concatenation failed\n");
 			return false;
@@ -566,8 +597,53 @@ bool codetree::parseConcatenation(xmldomnode *grammarnode,
 	return true;
 }
 
-bool codetree::parseAlternation(xmldomnode *grammarnode,
-				xmldomnode *treeparent,
+bool codetree::endOfStringOk(domnode *grammarnode) {
+
+	domnode	*def=NULL;
+	const char	*name=NULL;
+
+	// if we hit the end of the string, then we've only parsed sucessfully
+	// if this node and everything downstream of it is optional
+	for (domnode *sib=grammarnode;
+			!sib->isNullNode(); sib=sib->getNextTagSibling()) {
+
+		name=sib->getName();
+
+		if (!name) {
+			return false;
+		}
+
+		switch (*name) {
+			case OPTION:
+			case REPETITION:
+			case BREAK:
+			case EXCEPTION:
+				break;
+			case NONTERMINAL:
+				def=(domnode *)sib->getPrivateData();
+				if (!def || def->isNullNode()) {
+					return false;
+				}
+				if (!endOfStringOk(def->getFirstTagChild())) {
+					return false;
+				}
+				break;
+			case CONCATENATION:
+			case ALTERNATION:
+				if (!endOfStringOk(sib->getFirstTagChild())) {
+					return false;
+				}
+				break;
+			default:
+				return false;
+		}
+	}
+
+	return true;
+}
+
+bool codetree::parseAlternation(domnode *grammarnode,
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 	debugPrintIndent(4);
@@ -584,7 +660,7 @@ bool codetree::parseAlternation(xmldomnode *grammarnode,
 
 	// one of the children must parse successfully
 	bool	retval=false;
-	for (xmldomnode *child=grammarnode->getFirstTagChild();
+	for (domnode *child=grammarnode->getFirstTagChild();
 		!child->isNullNode(); child=child->getNextTagSibling()) {
 		if (parseChild(child,treeparent,codeposition,ntbuffer)) {
 			retval=true;
@@ -615,8 +691,8 @@ bool codetree::parseAlternation(xmldomnode *grammarnode,
 	return retval;
 }
 
-bool codetree::parseOption(xmldomnode *grammarnode, 
-				xmldomnode *treeparent,
+bool codetree::parseOption(domnode *grammarnode, 
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 	debugPrintIndent(4);
@@ -635,33 +711,22 @@ bool codetree::parseOption(xmldomnode *grammarnode,
 	return true;
 }
 
-bool codetree::parseRepetition(xmldomnode *grammarnode,
-				xmldomnode *treeparent,
+bool codetree::parseRepetition(domnode *grammarnode,
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 	debugPrintIndent(4);
 	debugPrintf(4,"repetition... {\n");
 
 	// there should be only one child, and zero or more instances of it
-	xmldomnode	*child=grammarnode->getFirstTagChild();
+	domnode	*child=grammarnode->getFirstTagChild();
 	bool		anyfound=false;
 	for (;;) {
 
-		// save old break state and reset break flag
-		//bool	oldbreak=pvt->_break;
-		//pvt->_break=false;
-
 		// parse the child
 		bool	parseresult=parseChild(child,treeparent,
 						codeposition,ntbuffer);
 
-		// process a break and restore break state
-		//if (pvt->_break) {
-			//parseresult=false;
-			//anyfound=true;
-		//}
-		//pvt->_break=oldbreak;
-
 		if (!parseresult || pvt->_endofstring) {
 			if (anyfound) {
 				debugPrintIndent(4);
@@ -676,8 +741,8 @@ bool codetree::parseRepetition(xmldomnode *grammarnode,
 	}
 }
 
-bool codetree::parseException(xmldomnode *grammarnode,
-				xmldomnode *treeparent,
+bool codetree::parseException(domnode *grammarnode,
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 	debugPrintIndent(4);
@@ -696,8 +761,8 @@ bool codetree::parseException(xmldomnode *grammarnode,
 	return false;
 }
 
-bool codetree::parseTerminal(xmldomnode *grammarnode,
-				xmldomnode *treeparent,
+bool codetree::parseTerminal(domnode *grammarnode,
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 
@@ -737,7 +802,7 @@ bool codetree::compareValue(const char *code,
 	// first, check for beginning-of-line characters,
 	// which should only occur at the beginning of a value
 	// (STX is used to represent the beginning of a line)
-	if (value && value[0]==STX) {
+	if (value && *value==STX) {
 		if (code==pvt->_beginningofinput || *(code-1)=='\n') {
 			value++;
 			(*valuelength)--;
@@ -746,14 +811,27 @@ bool codetree::compareValue(const char *code,
 		}
 	}
 
+	// see if we need to look for the end-of-transmission
+	bool	lookforeot=false;
+	size_t	lengthtocompare=*valuelength;
+	if (value && value[(*valuelength)-1]==EOT) {
+		lookforeot=true;
+		lengthtocompare--;
+	}
+
 	// see if the code matches this value
-	return (casesensitive && casesensitive[0]=='f')?
-		!charstring::compareIgnoringCase(value,code,*valuelength):
-		!charstring::compare(value,code,*valuelength);
+	if (!((casesensitive && *casesensitive=='f')?
+		!charstring::compareIgnoringCase(value,code,lengthtocompare):
+		!charstring::compare(value,code,lengthtocompare))) {
+		return false;
+	}
+
+	// if we must look for the end-of-transmission...
+	return (lookforeot)?(*(code+lengthtocompare)=='\0'):true;
 }
 
-bool codetree::parseLetter(xmldomnode *grammarnode,
-				xmldomnode *treeparent,
+bool codetree::parseLetter(domnode *grammarnode,
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 
@@ -779,8 +857,8 @@ bool codetree::parseLetter(xmldomnode *grammarnode,
 	return false;
 }
 
-bool codetree::parseLowerCaseLetter(xmldomnode *grammarnode,
-					xmldomnode *treeparent,
+bool codetree::parseLowerCaseLetter(domnode *grammarnode,
+					domnode *treeparent,
 					const char **codeposition,
 					stringbuffer *ntbuffer) {
 
@@ -807,8 +885,8 @@ bool codetree::parseLowerCaseLetter(xmldomnode *grammarnode,
 	return false;
 }
 
-bool codetree::parseUpperCaseLetter(xmldomnode *grammarnode,
-					xmldomnode *treeparent,
+bool codetree::parseUpperCaseLetter(domnode *grammarnode,
+					domnode *treeparent,
 					const char **codeposition,
 					stringbuffer *ntbuffer) {
 
@@ -835,8 +913,8 @@ bool codetree::parseUpperCaseLetter(xmldomnode *grammarnode,
 	return false;
 }
 
-bool codetree::parseDigit(xmldomnode *grammarnode,
-				xmldomnode *treeparent,
+bool codetree::parseDigit(domnode *grammarnode,
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 
@@ -862,8 +940,36 @@ bool codetree::parseDigit(xmldomnode *grammarnode,
 	return false;
 }
 
-bool codetree::parseSet(xmldomnode *grammarnode,
-				xmldomnode *treeparent,
+bool codetree::parseNonPrintableCharacter(
+				domnode *grammarnode,
+				domnode *treeparent,
+				const char **codeposition,
+				stringbuffer *ntbuffer) {
+
+	// if it matches, append the terminal to the
+	// nonterminal and advance the code position
+	if (!character::isPrintable((int32_t)(**codeposition))) {
+
+		// well... check for a break
+		if (parseBreakStack(codeposition)) {
+			return false;
+		}
+
+		debugPrintIndent(4);
+		debugPrintf(4,"digit found: \"%c\"\n",**codeposition);
+		if (ntbuffer) {
+			ntbuffer->append(**codeposition);
+		}
+		*codeposition=*codeposition+1;
+		return true;
+	}
+
+	// if it didn't match, don't advance the code position
+	return false;
+}
+
+bool codetree::parseSet(domnode *grammarnode,
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 
@@ -893,8 +999,8 @@ bool codetree::parseSet(xmldomnode *grammarnode,
 	return false;
 }
 
-bool codetree::parseBreak(xmldomnode *grammarnode,
-				xmldomnode *treeparent,
+bool codetree::parseBreak(domnode *grammarnode,
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 
@@ -905,7 +1011,7 @@ bool codetree::parseBreak(xmldomnode *grammarnode,
 	const char	*recurse=grammarnode->getAttributeValue(RECURSIVE);
 
 	// add this break to the stack, if necessary
-	if (recurse && recurse[0]=='t') {
+	if (recurse && *recurse=='t') {
 		break_t	*b=new break_t;
 		b->value=value;
 		b->valuelength=valuelength;
@@ -932,13 +1038,13 @@ bool codetree::parseBreak(xmldomnode *grammarnode,
 	return false;
 }
 
-bool codetree::parseNonTerminal(xmldomnode *grammarnode,
-				xmldomnode *treeparent,
+bool codetree::parseNonTerminal(domnode *grammarnode,
+				domnode *treeparent,
 				const char **codeposition,
 				stringbuffer *ntbuffer) {
 
-	// find the definition
-	xmldomnode	*def=(xmldomnode *)grammarnode->getPrivateData();
+	// get the definition
+	domnode	*def=(domnode *)grammarnode->getPrivateData();
 	if (!def || def->isNullNode()) {
 		debugPrintIndent(1);
 		debugPrintf(1,"ERROR: nonterminal %s not found\n",
@@ -946,18 +1052,28 @@ bool codetree::parseNonTerminal(xmldomnode *grammarnode,
 		return false;
 	}
 
-	// get the name
-	const char	*name=def->getAttributeValue(NAME);
-
 	// some variables...
-	xmldomnode	*codenode=NULL;
+	domnode	*codenode=NULL;
 	stringbuffer	*localntbuffer=NULL;
+	const char	*name=NULL;
+	const char	*alias=NULL;
+	char		symboltype='\0';
+
+	// we can defer getting the name/alias and symboltype until
+	// later unless we're debugging...
+	if (pvt->_debuglevel>=2) {
+		name=def->getAttributeValue(NAME);
+		alias=def->getAttributeValue(ALIAS);
+		if (alias) {
+			name=alias;
+		}
+		symboltype=getSymbolType(def);
+	}
 
 	// If a nonterminal buffer was passed in then we're apparently already
 	// building up another nonterminal that stores a literal value.  In
 	// that case we don't need to do anything but pass the current code
 	// parent and nonterminal buffer forward.
-	const char	symboltype=getSymbolType(def);
 	if (!ntbuffer) {
 
 		// If no nonterminal buffer was passed in...
@@ -969,10 +1085,19 @@ bool codetree::parseNonTerminal(xmldomnode *grammarnode,
 		// parent forward.  If it has a type of literal, then we
 		// need to create a buffer to build up the value in and pass it
 		// forward too...
+		symboltype=getSymbolType(def);
 		if (symboltype!=NONE) {
-			codenode=new xmldomnode(treeparent->getTree(),
+
+			// get the name/alias
+			name=def->getAttributeValue(NAME);
+			alias=def->getAttributeValue(ALIAS);
+			if (alias) {
+				name=alias;
+			}
+
+			codenode=new domnode(treeparent->getTree(),
 						treeparent->getNullNode(),
-						TAG_XMLDOMNODETYPE,
+						TAG_DOMNODETYPE,
 						pvt->_ns,name,NULL);
 		}
 		if (symboltype==LITERAL) {
@@ -1078,13 +1203,16 @@ bool codetree::parseBreakStack(const char **codeposition) {
 
 			// do the next few characters match?
 			break_t	*b=listnode->getValue();
+			const char	*value=b->value;
+			size_t		valuelength=b->valuelength;
+			const char	*casesensitive=b->casesensitive;
 			if (compareValue(*codeposition,
-					b->value,&(b->valuelength),
-					b->casesensitive)) {
+					value,&(valuelength),
+					casesensitive)) {
 
 				debugPrintIndent(2);
 				debugPrintf(2,"break \"");
-				debugSafePrint(2,b->value);
+				debugSafePrint(2,value);
 				debugPrintf(2,"\" found (from stack)\n");
 				pvt->_break=true;
 				return true;
@@ -1094,7 +1222,7 @@ bool codetree::parseBreakStack(const char **codeposition) {
 	return false;
 }
 
-bool codetree::write(xmldomnode *input,
+bool codetree::write(domnode *input,
 			const char *grammar,
 			stringbuffer *output) {
 
@@ -1117,7 +1245,7 @@ bool codetree::write(xmldomnode *input,
 	return write(input,&pvt->_grammar,output);
 }
 
-bool codetree::write(xmldomnode *input,
+bool codetree::write(domnode *input,
 			codetreegrammar *grammar,
 			stringbuffer *output) {
 
@@ -1144,7 +1272,7 @@ bool codetree::write(xmldomnode *input,
 	return writeNode(input,output);
 }
 
-bool codetree::writeNode(xmldomnode *node, stringbuffer *output) {
+bool codetree::writeNode(domnode *node, stringbuffer *output) {
 
 	// we're done
 	if (node->isNullNode()) {
@@ -1161,8 +1289,8 @@ bool codetree::writeNode(xmldomnode *node, stringbuffer *output) {
 	}
 
 	// if it's the root node then just write the children and return
-	if (node->getType()==ROOT_XMLDOMNODETYPE) {
-		for (xmldomnode *child=node->getFirstTagChild();
+	if (node->getType()==ROOT_DOMNODETYPE) {
+		for (domnode *child=node->getFirstTagChild();
 				!child->isNullNode();
 				child=child->getNextTagSibling()) {
 			if (!writeNode(child,output)) {
@@ -1173,9 +1301,8 @@ bool codetree::writeNode(xmldomnode *node, stringbuffer *output) {
 	}
 
 	// get the node's definition
-	xmldomnode	*def=pvt->_grammartag->getFirstTagChild(
-							DEFINITION,NAME,
-							node->getName());
+	domnode	*def=((codetreegrammar *)(pvt->_grammartag->getTree()))->
+					getDefinition(node->getName());
 	if (def->isNullNode()) {
 		debugPrintIndent(1);
 		debugPrintf(1,"ERROR: definition %s not found\n",
@@ -1203,7 +1330,7 @@ bool codetree::writeNode(xmldomnode *node, stringbuffer *output) {
 	writeStartEnd(output,start);
 	if (tag) {
 		output->append("<")->append(node->getName());
-		for (xmldomnode *att=node->getAttribute((uint64_t)0);
+		for (domnode *att=node->getAttribute((uint64_t)0);
 				!att->isNullNode(); att=att->getNextSibling()) {
 			if (charstring::compare(att->getName(),"value")) {
 				output->append(' ');
@@ -1233,7 +1360,7 @@ bool codetree::writeNode(xmldomnode *node, stringbuffer *output) {
 		}
 
 		// write the children
-		for (xmldomnode *child=node->getFirstTagChild();
+		for (domnode *child=node->getFirstTagChild();
 				!child->isNullNode();
 				child=child->getNextTagSibling()) {
 			if (!writeNode(child,output)) {
@@ -1288,7 +1415,7 @@ void codetree::writeStartEnd(stringbuffer *output, const char *string) {
 
 	// if it started with a backspace then either remove the preceeding
 	// carriage-return/line-feed or back up one level of indention
-	if (string[0]=='\b') {
+	if (*string=='\b') {
 
 		// make sure not to attempt to indent less than 0
 		if (pvt->_depth) {
