@@ -1,4 +1,4 @@
-// Copyright (c) 2002 David Muse
+// Copyright (c) 1999-2018 David Muse
 // See the COPYING file for more information
 
 
@@ -91,6 +91,14 @@
 	#define X_OK	0
 #endif
 
+#ifdef RUDIMENTS_HAVE_UNDEFINED_FSYNC
+extern "C" int fsync(int);
+#endif
+
+#ifdef RUDIMENTS_HAVE_UNDEFINED_FTRUNCATE
+extern "C" int ftruncate(int,int);
+#endif
+
 class fileprivate {
 	friend class file;
 	private:
@@ -155,14 +163,28 @@ void file::fileClone(const file &f) {
 }
 
 file::~file() {
+
+	// set NOTE in ~threadmutex()
+
+	if (!pvt) {
+		return;
+	}
+
 	#ifndef RUDIMENTS_HAVE_BLKSIZE_T
-		delete[] pvt->_name;
+		if (pvt->_name) {
+			char	*tmpname=pvt->_name;
+			pvt->_name=NULL;
+			delete[] tmpname;
+		}
 	#endif
-	delete pvt;
+
+	fileprivate	*tmppvt=pvt;
+	pvt=NULL;
+	delete tmppvt;
 }
 
 bool file::create(const char *name, mode_t perms) {
-	return open(name,O_CREAT|O_TRUNC|O_RDWR,perms);
+	return open(name,O_CREAT|O_TRUNC|O_RDWR|O_BINARY,perms);
 }
 
 bool file::createFile(const char *name, mode_t perms) {
@@ -173,6 +195,12 @@ bool file::createFile(const char *name, mode_t perms) {
 bool file::lowLevelOpen(const char *name, int32_t flags,
 				mode_t perms, bool useperms) {
 
+	// don't allow create without perms
+	if (flags&O_CREAT && (!perms || !useperms)) {
+		error::setErrorNumber(EINVAL);
+		return false;
+	}
+
 	#ifndef RUDIMENTS_HAVE_BLKSIZE_T
 		delete[] pvt->_name;
 		pvt->_name=charstring::duplicate(name);
@@ -360,30 +388,40 @@ char *file::getContents() {
 	if (fd()==-1) {
 		return NULL;
 	}
+	getCurrentProperties();
+	off64_t	curpos=getCurrentPosition();
+	setPositionRelativeToBeginning(0);
+	char	*contents=NULL;
 	if (pvt->_st.st_size) {
 		off64_t	size=pvt->_st.st_size;
-		char	*contents=new char[size+1];
+		contents=new char[size+1];
 		contents[size]='\0';
-		return (size==0 || read(contents,size)==size)?contents:NULL;
+		if (size && read(contents,size)!=size) {
+			delete[] contents;
+			contents=NULL;
+		}
 	} else {
-		stringbuffer	contents;
+		stringbuffer	cts;
 		// FIXME: this is fine for url's but for actual files that
 		// we don't know the size of, we ought to use a more
 		// intelligent size.
 		char		buffer[16*1024];
 		for (;;) {
 			ssize_t	s=read(buffer,sizeof(buffer));
-			contents.append(buffer,s);
+			cts.append(buffer,s);
 			if (s<(ssize_t)sizeof(buffer)) {
-				return contents.detachString();
+				contents=cts.detachString();
+				break;
 			}
 		}
 	}
+	setPositionRelativeToBeginning(curpos);
+	return contents;
 }
 
 char *file::getContents(const char *name) {
 	file	fl;
-	fl.open(name,O_RDONLY);
+	fl.open(name,O_RDONLY|O_BINARY);
 	char	*contents=fl.getContents();
 	fl.close();
 	return contents;
@@ -396,7 +434,7 @@ ssize_t file::getContents(unsigned char *buffer, size_t buffersize) {
 ssize_t file::getContents(const char *name, unsigned char *buffer,
 						size_t buffersize) {
 	file	fl;
-	fl.open(name,O_RDONLY);
+	fl.open(name,O_RDONLY|O_BINARY);
 	ssize_t	bytes=fl.getContents(buffer,buffersize);
 	fl.close();
 	return bytes;
@@ -674,16 +712,16 @@ bool file::truncate(const char *filename) {
 	return truncate(filename,0);
 }
 
-bool file::truncate(const char *filename, off64_t length) {
+bool file::truncate(const char *filename, off64_t size) {
 	#ifdef RUDIMENTS_HAVE_TRUNCATE
 		int32_t	result;
 		do {
-			result=::truncate(filename,length);
+			result=::truncate(filename,size);
 		} while (result==-1 && error::getErrorNumber()==-1);
 		return !result;
 	#else
 		file	f;
-		return f.open(filename,O_WRONLY) && f.truncate(length);
+		return f.open(filename,O_WRONLY) && f.truncate(size);
 	#endif
 }
 
@@ -691,13 +729,14 @@ bool file::truncate() const {
 	return truncate((off64_t)0);
 }
 
-bool file::truncate(off64_t length) const {
+bool file::truncate(off64_t size) const {
 	int32_t	result;
 	do {
-		#if defined(RUDIMENTS_HAVE_FTRUNCATE)
-			result=::ftruncate(fd(),length);
+		#if defined(RUDIMENTS_HAVE_FTRUNCATE) || \
+			defined(RUDIMENTS_HAVE_UNDEFINED_FTRUNCATE)
+			result=::ftruncate(fd(),size);
 		#elif defined(RUDIMENTS_HAVE__CHSIZE_S)
-			result=_chsize_s(fd(),length);
+			result=_chsize_s(fd(),size);
 		#elif defined(RUDIMENTS_HAVE_SETENDOFFILE)
 			return (setPositionRelativeToBeginning(0)!=-1 &&
 				SetEndOfFile((HANDLE)
@@ -770,7 +809,8 @@ bool file::accessible(const char *filename, int32_t mode) {
 	error::clearError();
 	do {
 		#if defined(RUDIMENTS_HAVE__ACCESS_S)
-			result=_access_s(filename,mode);
+			result=(!charstring::isNullOrEmpty(filename))?
+						_access_s(filename,mode):-1;
 		#elif defined(RUDIMENTS_HAVE_ACCESS)
 			result=access(filename,mode);
 		#else
@@ -796,7 +836,7 @@ bool file::getCurrentProperties() {
 	#ifndef RUDIMENTS_HAVE_BLKSIZE_T
 		filesystem	fs;
 		pvt->_blocksize=0;
-		if (fs.initialize(pvt->_name)) {
+		if (fs.open(pvt->_name)) {
 			pvt->_blocksize=fs.getBlockSize();
 		}
 	#endif
@@ -1145,19 +1185,6 @@ bool file::changeOwner(uid_t uid, gid_t gid) const {
 		if (!ue.initialize(uid) || !ge.initialize(gid)) {
 			return false;
 		}
-		// FIXME: do these need to be freed?
-		PSID	usid=(PSID)bytestring::duplicate(ue.getSid(),
-							ue.getSidSize());
-		PSID	gsid=(PSID)bytestring::duplicate(ge.getSid(),
-							ge.getSidSize());
-
-		// build trustees
-		TRUSTEE ut;
-		bytestring::zero(&ut,sizeof(ut));
-		BuildTrusteeWithSid(&ut,usid);
-		TRUSTEE gt;
-		bytestring::zero(&gt,sizeof(gt));
-		BuildTrusteeWithSid(&gt,gsid);
 
 		// adjust my privileges so I can set the owner
 		HANDLE	th=NULL;
@@ -1189,10 +1216,46 @@ bool file::changeOwner(uid_t uid, gid_t gid) const {
 			}
 		}
 
-		// build security descriptor
+		// set the owner/group
+		if (SetSecurityInfo(fh,
+				SE_FILE_OBJECT,
+				OWNER_SECURITY_INFORMATION|
+				GROUP_SECURITY_INFORMATION,
+				(PSID)ue.getSid(),
+				(PSID)ge.getSid(),
+				NULL,NULL)!=ERROR_SUCCESS) {
+			CloseHandle(th);
+			return false;
+		}
+
+		// Below is the original way I implemented this.
+		// BuildSecurityDescriptor crashes on Windows 8 though.
+		// There are other issues too.  Files are kernel objects, so I
+		// should probably use Get/SetKernelObjectSecurity instead of
+		// Get/SetUserObjectSecurity.  Though the docs also say to use
+		// SetSecurityInfo instead of SetKernelObjectSecurity for
+		// filesystem objects, which is what I changed it to.
+#if 0
+		// FIXME: do these need to be freed?
+		PSID	usid=(PSID)bytestring::duplicate(ue.getSid(),
+							ue.getSidSize());
+		PSID	gsid=(PSID)bytestring::duplicate(ge.getSid(),
+							ge.getSidSize());
+
+		// build trustees
+		TRUSTEE ut;
+		bytestring::zero(&ut,sizeof(ut));
+		BuildTrusteeWithSid(&ut,usid);
+		TRUSTEE gt;
+		bytestring::zero(&gt,sizeof(gt));
+		BuildTrusteeWithSid(&gt,gsid);
+
+		// get the existing owner/group security info for the file
+		// into a security descriptor
 		DWORD	sdsize=0;
 		SECURITY_INFORMATION	sinfo=
-			OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION;
+				OWNER_SECURITY_INFORMATION|
+				GROUP_SECURITY_INFORMATION;
 		GetUserObjectSecurity(fh,&sinfo,NULL,0,&sdsize);
 		PSECURITY_DESCRIPTOR	sdesc=
 			(PSECURITY_DESCRIPTOR)LocalAlloc(LPTR,sdsize);
@@ -1206,9 +1269,13 @@ bool file::changeOwner(uid_t uid, gid_t gid) const {
 			CloseHandle(th);
 			return false;
 		}
+
+		// build a new security descriptor, starting with the existing
+		// security descriptor, but updating the user/group info
 		PSECURITY_DESCRIPTOR	newsdesc=NULL;
 		DWORD	result=BuildSecurityDescriptor(
-						&ut,&gt,0,NULL,0,NULL,
+						&ut,&gt,
+						0,NULL,0,NULL,
 						sdesc,&sdsize,&newsdesc);
 		if (result!=ERROR_SUCCESS) {
 			LocalFree(sdesc);
@@ -1217,7 +1284,7 @@ bool file::changeOwner(uid_t uid, gid_t gid) const {
 			return false;
 		}
 
-		// set permissions
+		// apply the new security descriptor to the file
 		if (!SetUserObjectSecurity(fh,&sinfo,newsdesc)) {
 			LocalFree(sdesc);
 			LocalFree(newsdesc);
@@ -1225,8 +1292,10 @@ bool file::changeOwner(uid_t uid, gid_t gid) const {
 			return false;
 		}
 		
+		// clean up
 		LocalFree(sdesc);
 		LocalFree(newsdesc);
+#endif
 		CloseHandle(th);
 		return true;
 	#else
@@ -1372,7 +1441,8 @@ char *file::resolveSymbolicLink(const char *filename) {
 }
 
 bool file::sync() const {
-	#if defined(RUDIMENTS_HAVE_FSYNC)
+	#if defined(RUDIMENTS_HAVE_FSYNC) || \
+		defined(RUDIMENTS_HAVE_UNDEFINED_FSYNC)
 		int32_t	result;
 		error::clearError();
 		do {
@@ -1542,30 +1612,6 @@ bool file::createFifo(const char *filename, mode_t perms) {
 	#endif
 }
 
-bool file::createPipe(filedescriptor *readfd, filedescriptor *writefd) {
-	int32_t	result;
-	int	fd[2];
-	error::clearError();
-	do {
-		#if defined(RUDIMENTS_HAVE_PIPE)
-			result=pipe(fd);
-		#elif defined(RUDIMENTS_HAVE__PIPE)
-			result=_pipe(fd,1024,0);
-		#else
-			#error no pipe or anything like it
-		#endif
-	} while (result==-1 && error::getErrorNumber()==EINTR);
-	if (!result) {
-		if (readfd) {
-			readfd->setFileDescriptor(fd[0]);
-		}
-		if (writefd) {
-			writefd->setFileDescriptor(fd[1]);
-		}
-	}
-	return !result;
-}
-
 int32_t file::createTemporaryFile(char *templatefilename) {
 	#if defined(RUDIMENTS_HAVE_MKSTEMP)
 		int32_t	result;
@@ -1628,6 +1674,17 @@ int32_t file::createTemporaryFile(char *templatefilename) {
 	#endif
 }
 
+int32_t file::createTemporaryFile(char *templatefilename, mode_t perms) {
+	int32_t	result=createTemporaryFile(templatefilename);
+	if (result!=1) {
+		if (!permissions::setFilePermissions(result,perms)) {
+			remove(templatefilename);
+			return -1;
+		}
+	}
+	return result;
+}
+
 void *file::getInternalFileStatisticsStructure() {
 	return (void *)&(pvt->_st);
 }
@@ -1676,15 +1733,24 @@ char *file::basename(const char *filename) {
 	return retval;
 }
 
-char *file::basename(const char *filename, const char *suffix) {
+char *file::basename(const char *filename, const char *ext) {
 	char	*retval=basename(filename);
-	char	*ptr=charstring::findLast(retval,suffix);
-	if (!(*(ptr+charstring::length(suffix)))) {
+	char	*ptr=charstring::findLast(retval,ext);
+	if (!(*(ptr+charstring::length(ext)))) {
 		(*ptr)='\0';
 	}
 	return retval;
 }
 
+char *file::extension(const char *filename) {
+	const char	*ext="";
+	const char	*dot=charstring::findLast(filename,'.');
+	if (dot) {
+		ext=dot;
+	}
+	return charstring::duplicate(ext);
+}
+
 char *file::eightDotThree(const char *filename) {
 
 	// FIXME: I'm sure there's an infinitely more efficient way of
