@@ -1,4 +1,4 @@
-// Copyright (c) 2003 David Muse
+// Copyright (c) 1999-2018 David Muse
 // See the COPYING file for more information
 
 #include <rudiments/filesystem.h>
@@ -40,6 +40,10 @@
 	#ifdef RUDIMENTS_HAVE_WINDOWS_H
 		#include <windows.h>
 	#endif
+	#ifdef RUDIMENTS_HAVE_PSAPI_H
+		#define PSAPI_VERSION 1
+		#include <psapi.h>
+	#endif
 	#ifdef RUDIMENTS_HAVE_SYS_PARAM_H
 		#include <sys/param.h>
 	#endif
@@ -55,6 +59,13 @@
 	};
 #endif
 
+#ifdef RUDIMENTS_HAVE_ULTRIX_STATFS
+#include <sys/param.h>
+#include <sys/mount.h>
+extern "C" int statfs(char *,struct fs_data *);
+extern "C" int fstatfs(int,struct fs_data *);
+#endif
+
 class filesystemprivate {
 	friend class filesystem;
 	private:
@@ -62,7 +73,11 @@ class filesystemprivate {
 			struct		statvfs	_st;
 		#elif defined(RUDIMENTS_HAVE_SOME_KIND_OF_STATFS) || \
 				defined(RUDIMENTS_HAVE_WINDOWS_GETDISKFREESPACE)
-			struct		statfs	_st;
+			#ifdef RUDIMENTS_HAVE_ULTRIX_STATFS
+				struct		fs_data	_st;
+			#else
+				struct		statfs	_st;
+			#endif
 		#else
 			#error no statvfs, statfs or anything like it
 		#endif
@@ -110,7 +125,7 @@ filesystem::~filesystem() {
 	delete pvt;
 }
 
-bool filesystem::initialize(const char *path) {
+bool filesystem::open(const char *path) {
 	close();
 	if (charstring::isNullOrEmpty(path)) {
 		path=directory::getCurrentWorkingDirectory();
@@ -141,19 +156,21 @@ bool filesystem::initialize(const char *path) {
 	#endif
 }
 
-bool filesystem::initialize(int32_t fd) {
+bool filesystem::open(int32_t fd) {
 	close();
 	#ifndef RUDIMENTS_HAVE_WINDOWS_GETDISKFREESPACE
 		pvt->_fd=fd;
 		return getCurrentProperties();
 	#else
 		// This is complex on Windows.
+
 		// We have to determine the file name that the file descriptor
-		// refers to and call the other version of initialize().  On
-		// Vista and newer we can use GetFinalPathNameByHandle(), but
-		// on older versions we have to memory-map the file, get the
-		// file name from the map and convert the volume name to a
-		// dos-style drive letter.  We'll just go ahead and do that.
+		// refers to and call the other version of open().
+
+		// On Vista and newer we can use GetFinalPathNameByHandle(),
+		// but on older versions we have to memory-map the file, get
+		// the file name from the map and convert the volume name to a
+		// dos-style drive letter, so we'll just do that.
 
 		// get the file handle
 		HANDLE	fh=(HANDLE)filedescriptor::
@@ -182,21 +199,12 @@ bool filesystem::initialize(int32_t fd) {
 			return false;
 		}
 
-		// get GetMappedFileName function
-		HMODULE	kernel32=GetModuleHandle("Kernel32");
-		if (!kernel32) {
-			return false;
-		}
-		DWORD (*getmappedfilename)(HANDLE, LPVOID, LPTSTR, DWORD)=
-			(DWORD (*)(HANDLE, LPVOID, LPTSTR, DWORD))
-			GetProcAddress(kernel32,"K32GetMappedFileNameA");
-		if (!getmappedfilename) {
-			return false;
-		}
-
 		// get file name
+		// NOTE: If GetMappedFileName is dynamically loaded on
+		// Windows 8 or higher then it corrupts the stack when it is
+		// used.  It works fine on Windows 7 and lower though.
 		char	filename[MAX_PATH+1];
-		if (!getmappedfilename(GetCurrentProcess(),
+		if (!GetMappedFileName(GetCurrentProcess(),
 					fv,filename,MAX_PATH)) {
 			UnmapViewOfFile(fv);
 			CloseHandle(fm);
@@ -217,10 +225,10 @@ bool filesystem::initialize(int32_t fd) {
 		volume[0]='\0';
 
 		// set up a buffer for the drive mapping
-		char	mapping[MAX_PATH];
+		char	mapping[MAX_PATH+1];
 
 		// for each volume...
-		for (char driveletter='A'; driveletter<='Z'; driveletter++) {
+		for (char driveletter='c'; driveletter<='z'; driveletter++) {
 
 			// ignore volumes that don't exist
 			DWORD	exists=volumes&0x0001;
@@ -233,7 +241,14 @@ bool filesystem::initialize(int32_t fd) {
 			volume[0]=driveletter;
 
 			// get the drive mapping
-			if (!QueryDosDevice(volume,mapping,sizeof(mapping))) {
+			// FIXME: If the volume is mapped to more than 1
+			// location then this could return 0 with
+			// ERROR_INSUFFICIENT_BUFFER if mapping[] is too small
+			// to store the list of locations.  In that case we
+			// should loop back, resize mapping[] and try again.
+			DWORD	result=QueryDosDevice(volume,mapping,
+							MAX_PATH/sizeof(TCHAR));
+			if (!result) {
 				UnmapViewOfFile(fv);
 				CloseHandle(fm);
 				return false;
@@ -241,6 +256,9 @@ bool filesystem::initialize(int32_t fd) {
 
 			// if the drive mapping for this volume matches
 			// then we're done
+			// FIXME: If the volume is mapped to more than 1
+			// location then we need to test the filename against
+			// each of the mappings.
 			if (!charstring::compare(filename,mapping,
 						charstring::length(mapping))) {
 				break;
@@ -263,8 +281,8 @@ bool filesystem::initialize(int32_t fd) {
 			*(ptr+1)=':';
 		}
 		
-		// initialize using file name
-		return (volume[0])?initialize(ptr):false;
+		// open using file name
+		return (volume[0])?open(ptr):false;
 	#endif
 }
 
@@ -376,6 +394,7 @@ bool filesystem::getCurrentProperties() {
 }
 
 int64_t filesystem::getType() const {
+	int64_t	result=0;
 #if defined(RUDIMENTS_HAVE_LINUX_STATFS) || \
 	defined(RUDIMENTS_HAVE_LINUX_LIBC4_STATFS) || \
 	defined(RUDIMENTS_HAVE_FREEBSD_STATFS) || \
@@ -383,10 +402,157 @@ int64_t filesystem::getType() const {
 	defined(RUDIMENTS_HAVE_DARWIN_STATFS) || \
 	defined(RUDIMENTS_HAVE_CYGWIN_STATFS) || \
 	defined(RUDIMENTS_HAVE_WINDOWS_GETDISKFREESPACE)
-	return pvt->_st.f_type;
-#else
-	return 0;
+	result=pvt->_st.f_type;
 #endif
+	if (result) {
+		return result;
+	}
+	const char	*fstypename=NULL;
+#if defined(RUDIMENTS_HAVE_STATVFS)
+	fstypename=(const char *)pvt->_st.f_basetype;
+#elif defined(RUDIMENTS_HAVE_NETBSD_STATVFS) || \
+	defined(RUDIMENTS_HAVE_OPENBSD_STATFS)
+	fstypename=(const char *)pvt->_st.f_fstypename;
+#endif
+	if (!charstring::compareIgnoringCase(
+				fstypename,"adfs")) {
+		return 0xADF5;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"affs")) {
+		return 0xADFF;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"befs")) {
+		return 0x42465331;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"bfs")) {
+		return 0x1BADFACE;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"cifs")) {
+		return 0xFF534D42;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"codafs")) {
+		return 0x73757245;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"coherent")) {
+		return 0x012FF7B7;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"cramfs")) {
+		return 0x28cd3d45;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"devfs")) {
+		return 0x1373;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"efs")) {
+		return 0x00414A53;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"ext")) {
+		return 0x137D;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"ext2 (old)")) {
+		return 0xEF51;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"ext2")) {
+		return 0xEF53;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"hfs")) {
+		return 0x4244;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"hpfs")) {
+		return 0xF995E849;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"htfs")) {
+		return 0xA060492A;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"hugetlbfs")) {
+		return 0x958458f6;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"isofs")) {
+		return 0x9660;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"jffs2")) {
+		return 0x72b6;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"jfs")) {
+		return 0x3153464a;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"minix (original)")) {
+		return 0x137F;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"minix (30 char)")) {
+		return 0x138F;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"minix2 (original)")) {
+		return 0x2468;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"minix2 (30 char)")) {
+		return 0x2478;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"msdos")) {
+		return 0x4d44;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"ncp")) {
+		return 0x564c;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"nfs")) {
+		return 0x6969;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"ntfs")) {
+		return 0x5346544e;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"openprom")) {
+		return 0x9fa1;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"proc")) {
+		return 0x9fa0;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"qnx4")) {
+		return 0x002f;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"reiserfs")) {
+		return 0x52654973;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"romfs")) {
+		return 0x7275;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"smb")) {
+		return 0x517B;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"sysv2")) {
+		return 0x012FF7B6;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"sysv4")) {
+		return 0x012FF7B5;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"tmpfs")) {
+		return 0x01021994;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"udf")) {
+		return 0x15013346;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"ufs") ||
+			!charstring::compareIgnoringCase(
+				fstypename,"ffs")) {
+		return 0x00011954;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"usbdevfs")) {
+		return 0x9fa2;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"vxfs")) {
+		return 0xa501FCF5;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"xenix")) {
+		return 0x012FF7B4;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"xfs")) {
+		return 0x58465342;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"xiafs")) {
+		return 0x012FD16D;
+	} else if (!charstring::compareIgnoringCase(
+				fstypename,"zfs")) {
+		return 0x00BAB10C;
+	}
+	return 0;
 }
 
 int64_t filesystem::getBlockSize() const {
@@ -478,6 +644,14 @@ int64_t filesystem::getAvailableBlocks() const {
 #endif
 }
 
+int64_t filesystem::getReservedBlocks() const {
+#if defined(RUDIMENTS_HAVE_STATVFS_RESVD)
+	return pvt->_st.f_bresvd;
+#else
+	return 0;
+#endif
+}
+
 int64_t filesystem::getTotalFileNodes() const {
 #if defined(RUDIMENTS_HAVE_LINUX_STATFS) || \
 	defined(RUDIMENTS_HAVE_LINUX_LIBC4_STATFS) || \
@@ -516,14 +690,40 @@ int64_t filesystem::getAvailableFileNodes() const {
 #if defined(RUDIMENTS_HAVE_STATVFS) || \
 	defined(RUDIMENTS_HAVE_NETBSD_STATVFS) || \
 	defined(RUDIMENTS_HAVE_MINIX_HAIKU_STATVFS)
-	return pvt->_st.f_ffree;
+	return pvt->_st.f_favail;
+#else
+	return 0;
+#endif
+}
+
+int64_t filesystem::getReservedFileNodes() const {
+#if defined(RUDIMENTS_HAVE_STATVFS_RESVD)
+	return pvt->_st.f_fresvd;
 #else
 	return 0;
 #endif
 }
 
 int64_t filesystem::getFileSystemId() const {
-#if defined(RUDIMENTS_HAVE_STATVFS) || \
+#if defined(RUDIMENTS_HAVE_LINUX_STATFS)
+	// fsid_t is an opaque struct, but it typically contains an array of
+	// some number of bytes which are collectively set to some unique value.
+	// Attempt to handle various possibilities...
+	if (sizeof(fsid_t)==sizeof(uint16_t)) {
+		uint16_t	temp;
+		bytestring::copy(&temp,&pvt->_st.f_fsid,sizeof(uint16_t));
+		return (int64_t)temp;
+	} else if (sizeof(fsid_t)==sizeof(uint32_t)) {
+		uint32_t	temp;
+		bytestring::copy(&temp,&pvt->_st.f_fsid,sizeof(uint32_t));
+		return (int64_t)temp;
+	} else {
+		uint64_t	temp;
+		bytestring::copy(&temp,&pvt->_st.f_fsid,sizeof(uint64_t));
+		return (int64_t)temp;
+	}
+	return 0;
+#elif defined(RUDIMENTS_HAVE_STATVFS) || \
 	defined(RUDIMENTS_HAVE_NETBSD_STATVFS) || \
 	defined(RUDIMENTS_HAVE_MINIX_HAIKU_STATVFS) || \
 	defined(RUDIMENTS_HAVE_CYGWIN_STATFS) || \
@@ -589,9 +789,11 @@ const char *filesystem::getMountPoint() const {
 	defined(RUDIMENTS_HAVE_NETBSD_STATVFS) || \
 	defined(RUDIMENTS_HAVE_OPENBSD_STATFS) || \
 	defined(RUDIMENTS_HAVE_DARWIN_STATFS)
-	return (const char *)pvt->_st.f_mntonname;
+	return (pvt->_st.f_mntonname[0])?
+			(const char *)pvt->_st.f_mntonname:NULL;
 #elif defined(RUDIMENTS_HAVE_WINDOWS_GETDISKFREESPACE)
-	return (const char *)pvt->_volume;
+	return (!charstring::isNullOrEmpty(pvt->_volume))?
+				(const char *)pvt->_volume:NULL;
 #else
 	return NULL;
 #endif
@@ -620,7 +822,8 @@ const char *filesystem::getDeviceName() const {
 	defined(RUDIMENTS_HAVE_NETBSD_STATVFS) || \
 	defined(RUDIMENTS_HAVE_OPENBSD_STATFS) || \
 	defined(RUDIMENTS_HAVE_DARWIN_STATFS)
-	return (const char *)pvt->_st.f_mntfromname;
+	return (pvt->_st.f_mntfromname[0])?
+			(const char *)pvt->_st.f_mntfromname:NULL;
 #else
 	return NULL;
 #endif
@@ -628,9 +831,11 @@ const char *filesystem::getDeviceName() const {
 
 const char *filesystem::getFilesystemSpecificString() const {
 #ifdef RUDIMENTS_HAVE_STATVFS
-	return (const char *)pvt->_st.f_fstr;
+	return (pvt->_st.f_fstr[0])?
+			(const char *)pvt->_st.f_fstr:NULL;
 #elif defined(RUDIMENTS_HAVE_WINDOWS_GETDISKFREESPACE)
-	return (const char *)pvt->_st.f_volumename;
+	return (pvt->_st.f_volumename[0])?
+			(const char *)pvt->_st.f_volumename:NULL;
 #else
 	return NULL;
 #endif
@@ -646,12 +851,21 @@ const char *filesystem::getTypeName() const {
 	defined(RUDIMENTS_HAVE_NETBSD_STATVFS) || \
 	defined(RUDIMENTS_HAVE_OPENBSD_STATFS) || \
 	defined(RUDIMENTS_HAVE_DARWIN_STATFS)
-	return (const char *)pvt->_st.f_fstypename;
+	return (pvt->_st.f_fstypename[0])?
+			(const char *)pvt->_st.f_fstypename:NULL;
 #elif defined(RUDIMENTS_HAVE_STATVFS)
-	return (const char *)pvt->_st.f_basetype;
+	return (pvt->_st.f_basetype[0])?
+			(const char *)pvt->_st.f_basetype:NULL;
 #elif defined(RUDIMENTS_HAVE_LINUX_STATFS) || \
 	defined(RUDIMENTS_HAVE_LINUX_LIBC4_STATFS)
-	switch (pvt->_st.f_type) {
+	// NOTE: casting pvt->_st.f_type to uint64_t in the switch below
+	// resolves issues on systems where it's defined as a signed int
+	// and comparisons with values like 0xFF534D42 would throw a
+	// "narrowing" warning as those values are larger than a signed int
+	// can hold.  I'm casting it to a uint64_t rather than uint32_t
+	// because, who knows, it might be 64-bits on some system and I
+	// don't want to down-cast it to 32-bits.
+	switch ((uint64_t)pvt->_st.f_type) {
 		case 0xADF5:
 			return "adfs";
 		case 0xADFF:
@@ -682,6 +896,8 @@ const char *filesystem::getTypeName() const {
 			return "hfs";
 		case 0xF995E849:
 			return "hpfs";
+		case 0xA060492A:
+			return "htfs";
 		case 0x958458f6:
 			return "hugetlbfs";
 		case 0x9660:
@@ -738,10 +954,13 @@ const char *filesystem::getTypeName() const {
 			return "xfs";
 		case 0x012FD16D:
 			return "xiafs";
+		case 0x00BAB10C:
+			return "zfs";
 	}
 	return NULL;
 #elif defined(RUDIMENTS_HAVE_WINDOWS_GETDISKFREESPACE)
-	return (const char *)pvt->_st.f_typename;
+	return (pvt->_st.f_typename[0])?
+			(const char *)pvt->_st.f_typename:NULL;
 #else
 	return NULL;
 #endif
