@@ -1,4 +1,4 @@
-// Copyright (c) 2015 David Muse
+// Copyright (c) 1999-2018 David Muse
 // See the COPYING file for more information.
 
 #include <rudiments/charstring.h>
@@ -160,16 +160,6 @@ void gss::clear() {
 	}
 }
 
-bool gss::supportsGSS() {
-	#if defined(RUDIMENTS_HAS_GSS)
-		return true;
-	#elif defined(RUDIMENTS_HAS_SSPI)
-		return true;
-	#else
-		return false;
-	#endif
-}
-
 const char *gss::getSspiStatusString(uint32_t status) {
 	#if defined(RUDIMENTS_HAS_SSPI)
 		switch (status) {
@@ -359,12 +349,16 @@ const char *gss::getSspiStatusString(uint32_t status) {
 			case SEC_E_NO_KERB_KEY:
 				return "SEC_E_NO_KERB_KEY";
 				break;
+			#ifdef SEC_E_CERT_WRONG_USAGE
 			case SEC_E_CERT_WRONG_USAGE:
 				return "SEC_E_CERT_WRONG_USAGE";
 				break;
+			#endif
+			#ifdef SEC_E_DOWNGRADE_DETECTED
 			case SEC_E_DOWNGRADE_DETECTED:
 				return "SEC_E_DOWNGRADE_DETECTED";
 				break;
+			#endif
 			case SEC_E_SMARTCARD_CERT_REVOKED:
 				return "SEC_E_SMARTCARD_CERT_REVOKED";
 				break;
@@ -380,48 +374,76 @@ const char *gss::getSspiStatusString(uint32_t status) {
 			case SEC_E_SMARTCARD_CERT_EXPIRED:
 				return "SEC_E_SMARTCARD_CERT_EXPIRED";
 				break;
+			#ifdef SEC_E_NO_S4U_PROT_SUPPORT
 			case SEC_E_NO_S4U_PROT_SUPPORT:
 				return "SEC_E_NO_S4U_PROT_SUPPORT";
 				break;
+			#endif
+			#ifdef SEC_E_CROSSREALM_DELEGATION_FAILURE
 			case SEC_E_CROSSREALM_DELEGATION_FAILURE:
 				return "SEC_E_CROSSREALM_DELEGATION_FAILURE";
 				break;
+			#endif
+			#ifdef SEC_E_REVOCATION_OFFLINE_KDC
 			case SEC_E_REVOCATION_OFFLINE_KDC:
 				return "SEC_E_REVOCATION_OFFLINE_KDC";
 				break;
+			#endif
+			#ifdef SEC_E_ISSUING_CA_UNTRUSTED_KDC
 			case SEC_E_ISSUING_CA_UNTRUSTED_KDC:
 				return "SEC_E_ISSUING_CA_UNTRUSTED_KDC";
 				break;
+			#endif
+			#ifdef SEC_E_KDC_CERT_EXPIRED
 			case SEC_E_KDC_CERT_EXPIRED:
 				return "SEC_E_KDC_CERT_EXPIRED";
 				break;
+			#endif
+			#ifdef SEC_E_KDC_CERT_REVOKED
 			case SEC_E_KDC_CERT_REVOKED:
 				return "SEC_E_KDC_CERT_REVOKED";
 				break;
+			#endif
+			#ifdef SEC_I_SIGNATURE_NEEDED
 			case SEC_I_SIGNATURE_NEEDED:
 				return "SEC_I_SIGNATURE_NEEDED";
 				break;
+			#endif
+			#ifdef SEC_E_INVALID_PARAMETER
 			case SEC_E_INVALID_PARAMETER:
 				return "SEC_E_INVALID_PARAMETER";
 				break;
+			#endif
+			#ifdef SEC_E_DELEGATION_POLICY
 			case SEC_E_DELEGATION_POLICY:
 				return "SEC_E_DELEGATION_POLICY";
 				break;
+			#endif
+			#ifdef SEC_E_POLICY_NLTM_ONLY
 			case SEC_E_POLICY_NLTM_ONLY:
 				return "SEC_E_POLICY_NLTM_ONLY";
 				break;
+			#endif
+			#ifdef SEC_I_NO_RENEGOTIATION
 			case SEC_I_NO_RENEGOTIATION:
 				return "SEC_I_NO_RENEGOTIATION";
 				break;
+			#endif
+			#ifdef SEC_E_NO_CONTEXT
 			case SEC_E_NO_CONTEXT:
 				return "SEC_E_NO_CONTEXT";
 				break;
+			#endif
+			#ifdef SEC_E_PKU2U_CERT_FAILURE
 			case SEC_E_PKU2U_CERT_FAILURE:
 				return "SEC_E_PKU2U_CERT_FAILURE";
 				break;
+			#endif
+			#ifdef SEC_E_MUTUAL_AUTH_FAILED
 			case SEC_E_MUTUAL_AUTH_FAILED:
 				return "SEC_E_MUTUAL_AUTH_FAILED";
 				break;
+			#endif
 			case SEC_E_OK:
 				return "SEC_E_OK";
 				break;
@@ -433,6 +455,16 @@ const char *gss::getSspiStatusString(uint32_t status) {
 	#endif
 }
 
+bool gss::supported() {
+	#if defined(RUDIMENTS_HAS_GSS)
+		return true;
+	#elif defined(RUDIMENTS_HAS_SSPI)
+		return true;
+	#else
+		return false;
+	#endif
+}
+
 
 class gssmechanismprivate {
 	friend class gssmechanism;
@@ -1100,12 +1132,7 @@ void gsscredentials::close() {
 			pvt->_actualmechanisms=GSS_C_NO_OID_SET;
 		}
 	#endif
-	for (linkedlistnode< gssmechanism * > *node=
-					pvt->_amlist.getFirst();
-					node; node=node->getNext()) {
-		delete node->getValue();
-	}
-	pvt->_amlist.clear();
+	pvt->_amlist.clearAndDelete();
 	#if defined(RUDIMENTS_HAS_GSS)
 		pvt->_actuallifetime=0;
 	#elif defined(RUDIMENTS_HAS_SSPI)
@@ -3043,11 +3070,10 @@ bool gsscontext::wrap(const unsigned char *input,
 			*output=new BYTE[*outputsize];
 
 			// store that data, in that order
-			*((DWORD *)*output)=inputsize;
-			bytestring::copy(*output+sizeof(DWORD),
-						input,inputsize);
-			*((DWORD *)(*output+sizeof(DWORD)+inputsize))=
-							pvt->_trailersize;
+			bytestring::copy(*output,&inputsize,sizeof(DWORD));
+			bytestring::copy(*output+sizeof(DWORD),input,inputsize);
+			bytestring::copy(*output+sizeof(DWORD)+inputsize,
+					&pvt->_trailersize,sizeof(DWORD));
 
 			// prepare security buffers
 			SecBuffer         secbuf[2];
@@ -3264,7 +3290,7 @@ bool gsscontext::unwrap(const unsigned char *input,
 			// "data" buffer.
 
 			// get output size, output, trailer size, and trailer
-			*outputsize=*((DWORD *)input);
+			bytestring::copy(outputsize,input,sizeof(DWORD));
 
 			// create a buffer to store the output and initialize
 			// it with the contents of the "data" buffer
@@ -3273,8 +3299,10 @@ bool gsscontext::unwrap(const unsigned char *input,
 							*outputsize);
 
 			// get the trailer size and location of the trailer
-			DWORD	trailersize=
-				*((DWORD *)(input+sizeof(DWORD)+*outputsize));
+			DWORD	trailersize;
+			bytestring::copy(&trailersize,
+				(DWORD *)(input+sizeof(DWORD)+*outputsize),
+				sizeof(DWORD));
 			BYTE	*trailer=(BYTE *)
 				(input+sizeof(DWORD)+*outputsize+sizeof(DWORD));
 
@@ -3411,10 +3439,10 @@ bool gsscontext::verifyMic(const unsigned char *message,
 	return false;
 }
 
-ssize_t gsscontext::read(void *buf, ssize_t count) {
+ssize_t gsscontext::read(void *buf, ssize_t size) {
 
 	// first, return any buffered data
-	ssize_t bytestoread=count;
+	ssize_t bytestoread=size;
 	ssize_t	bytesread=0;
 	size_t	pendingbytes=pending();
 	if (pendingbytes) {
@@ -3444,7 +3472,7 @@ ssize_t gsscontext::read(void *buf, ssize_t count) {
 	//
 	// This should be a low-level read, and it should allow short reads.
 	// I.e. it should't try to receive multiple tokens until it's read
-	// "count" bytes.  It should just return "what's currently available",
+	// "size" bytes.  It should just return "what's currently available",
 	// and in this case that should come from whatever was already buffered,
 	// and/or the next token.  The filedescriptor class's read-buffering
 	// depends on this method having short-read behavior.  We'll leave it
@@ -3604,10 +3632,6 @@ ssize_t gsscontext::receiveTlsToken(uint32_t *tokenflags,
 	}
 
 	// get record size
-	// (some compiler options don't allow setting it directly:
-	// error: dereferencing type-punned pointer will break
-	//        strict-aliasing rules [-Werror=strict-aliasing])
-	//uint16_t	size=*((uint16_t *)&(header[3]));
 	uint16_t	size=0;
 	bytestring::copy(&size,&(header[3]),sizeof(uint16_t));
 	size=pvt->_fd->netToHost(size);
@@ -3661,12 +3685,12 @@ ssize_t gsscontext::receiveTlsToken(uint32_t *tokenflags,
 	return *tokensize;
 }
 
-ssize_t gsscontext::write(const void *buf, ssize_t count) {
+ssize_t gsscontext::write(const void *buf, ssize_t size) {
 
 	// create token
 	unsigned char	*tokendata=NULL;
 	size_t		tokensize=0;
-	if (!wrap((const unsigned char *)buf,count,&tokendata,&tokensize)) {
+	if (!wrap((const unsigned char *)buf,size,&tokendata,&tokensize)) {
 		delete[] tokendata;
 		return RESULT_ERROR;
 	}
@@ -3677,7 +3701,7 @@ ssize_t gsscontext::write(const void *buf, ssize_t count) {
 	if (result!=(ssize_t)tokensize) {
 		return (result<=0)?result:RESULT_ERROR;
 	}
-	return count;
+	return size;
 }
 
 ssize_t gsscontext::sendToken(uint32_t tokenflags,
