@@ -1,4 +1,4 @@
-// Copyright (c) 2002 David Muse
+// Copyright (c) 1999-2018 David Muse
 // See the COPYING file for more information
 
 #include <rudiments/listener.h>
@@ -123,12 +123,16 @@ void listener::cleanUp() {
 		}
 		delete[] pvt->_kevs;
 		delete[] pvt->_rkevs;
+		pvt->_kevs=NULL;
+		pvt->_rkevs=NULL;
 	#elif defined(RUDIMENTS_HAVE_EPOLL)
 		if (pvt->_epfd!=-1) {
 			::close(pvt->_epfd);
 		}
 		delete[] pvt->_evs;
 		delete[] pvt->_revs;
+		pvt->_evs=NULL;
+		pvt->_revs=NULL;
 	#elif defined(RUDIMENTS_HAVE_PORT_CREATE)
 		if (pvt->_port!=-1) {
 			::close(pvt->_port);
@@ -136,8 +140,10 @@ void listener::cleanUp() {
 	#elif defined(RUDIMENTS_HAVE_SYS_DEVPOLL_H)
 		pvt->_dpfd.close();
 		delete[] pvt->_fds;
+		pvt->_fds=NULL;
 	#elif defined(RUDIMENTS_HAVE_POLL)
 		delete[] pvt->_fds;
+		pvt->_fds=NULL;
 	#elif defined(RUDIMENTS_HAVE_CREATE_IO_COMPLETION_PORT)
 		CloseHandle(pvt->_cp);
 	#endif
@@ -187,11 +193,7 @@ void listener::removeFileDescriptor(filedescriptor *fd) {
 }
 
 void listener::removeAllFileDescriptors() {
-	for (linkedlistnode< fddata_t * > *node=pvt->_fdlist.getFirst();
-						node; node=node->getNext()) {
-		delete node->getValue();
-	}
-	pvt->_fdlist.clear();
+	pvt->_fdlist.clearAndDelete();
 	pvt->_dirty=true;
 }
 
@@ -300,21 +302,54 @@ int32_t listener::listen(int32_t sec, int32_t usec) {
 		#if defined(RUDIMENTS_HAVE_KQUEUE)
 
 			// wait for non-blocking io
-			result=kevent(pvt->_kq,pvt->_kevs,fdcount,
+			bool	loop=true;
+			while (loop) {
+				result=kevent(pvt->_kq,
+						pvt->_kevs,fdcount,
 						pvt->_rkevs,fdcount,
 						tsptr);
+				if (!result) {
+					loop=false;
+				}
+				for (int32_t i=0; i<result; i++) {
+					if (pvt->_rkevs[i].filter==
+							EVFILT_READ ||
+						pvt->_rkevs[i].filter==
+							EVFILT_WRITE) {
+						loop=false;
+						break;
+					}
+				}
+			}
 
 		#elif defined(RUDIMENTS_HAVE_EPOLL)
 
 			// wait for non-blocking io
-			result=epoll_wait(pvt->_epfd,
-						pvt->_revs,fdcount,
-						timeout);
+			bool	loop=true;
+			while (loop) {
+				result=epoll_wait(pvt->_epfd,
+							pvt->_revs,fdcount,
+							timeout);
+				if (!result) {
+					loop=false;
+				}
+				for (int32_t i=0; i<result; i++) {
+					if (pvt->_revs[i].events&
+							(EPOLLIN|EPOLLOUT)) {
+						loop=false;
+						break;
+					}
+				}
+			}
 
 		#elif defined(RUDIMENTS_HAVE_PORT_CREATE)
 
-			// wait for an event
-			result=port_get(pvt->_port,&pvt->_pev,tsptr);
+			// wait for non-blocking io
+			pvt->_pev.portev_events=0;
+			do {
+				result=port_get(pvt->_port,&pvt->_pev,tsptr);
+			} while (!(pvt->_pev.portev_events&(POLLIN|POLLOUT)) &&
+					error::getErrorNumber()!=ETIME);
 			if (!result) {
 				result=1;
 			} else if (result==-1 &&
