@@ -1,9 +1,10 @@
-// Copyright (c) 1999-2002 David Muse
+// Copyright (c) 1999-2018 David Muse
 // See the COPYING file for more information
 
 #include <rudiments/semaphoreset.h>
 #include <rudiments/userentry.h>
 #include <rudiments/groupentry.h>
+#include <rudiments/stdio.h>
 #include <rudiments/error.h>
 
 #ifdef RUDIMENTS_HAVE_CREATESEMAPHORE
@@ -61,6 +62,7 @@ class semaphoresetprivate {
 			char	**_semnames;
 			HMODULE	_lib;
 		#endif
+		bool	_supportstimedops;
 };
 
 semaphoreset::semaphoreset() {
@@ -81,6 +83,26 @@ semaphoreset::semaphoreset() {
 		pvt->_semnames=NULL;
 		pvt->_lib=NULL;
 	#endif
+	#if defined(RUDIMENTS_HAVE_SEMTIMEDOP)
+		// We can't implicitly trust RUDIMENTS_HAVE_SEMTIMEDOP because
+		// some platforms (eg. older sparc linux) implement semtimedop
+		// to just set ENOSYS and return -1.  Fortunately, these
+		// platforms do this even if you pass it garbage parameters,
+		// and other platforms set some other error and return -1 when
+		// you pass it garbage parameters.  So, we can do an inexpensive
+		// and non-invasive test here to see if we have a working
+		// semtimedop() or not.
+		int32_t	result;
+		error::clearError();
+		do {
+			result=semtimedop(-1,NULL,0,NULL);
+		} while (result==-1 && error::getErrorNumber()==EINTR);
+		pvt->_supportstimedops=(error::getErrorNumber()!=ENOSYS);
+	#elif defined(RUDIMENTS_HAVE_CREATESEMAPHORE)
+		pvt->_supportstimedops=true;
+	#else
+		pvt->_supportstimedops=false;
+	#endif
 }
 
 semaphoreset::~semaphoreset() {
@@ -568,13 +590,22 @@ int32_t semaphoreset::semGet(key_t key, int32_t nsems,
 		for (int32_t i=0; i<nsems; i++) {
 
 			// set the semaphore name
-			int32_t	semnamelen=11+charstring::integerLength(key)+1+
-						charstring::integerLength(i)+1;
-			pvt->_semnames[i]=new char[semnamelen];
-			charstring::copy(pvt->_semnames[i],"rudiments::");
-			charstring::append(pvt->_semnames[i],(int64_t)key);
-			charstring::append(pvt->_semnames[i],"-");
-			charstring::append(pvt->_semnames[i],(int64_t)i);
+			if (key!=IPC_PRIVATE) {
+				int32_t	semnamelen=11+
+					charstring::integerLength(key)+1+
+					charstring::integerLength(i)+1;
+				pvt->_semnames[i]=new char[semnamelen];
+				charstring::copy(
+					pvt->_semnames[i],"rudiments::");
+				charstring::append(
+					pvt->_semnames[i],(int64_t)key);
+				charstring::append(
+					pvt->_semnames[i],"-");
+				charstring::append(
+					pvt->_semnames[i],(int64_t)i);
+			} else {
+				pvt->_semnames[i]=NULL;
+			}
 
 			if (semflg&(IPC_CREAT|IPC_EXCL)) {
 
@@ -691,12 +722,7 @@ bool semaphoreset::semTimedOp(struct sembuf *sops,
 }
 
 bool semaphoreset::supportsTimedSemaphoreOperations() {
-	#if defined(RUDIMENTS_HAVE_SEMTIMEDOP) || \
-		defined(RUDIMENTS_HAVE_CREATESEMAPHORE)
-		return true;
-	#else
-		return false;
-	#endif
+	return pvt->_supportstimedops;
 }
 
 bool semaphoreset::supportsUndoSemaphoreOperations() {
@@ -714,3 +740,19 @@ void semaphoreset::retryInterruptedOperations() {
 void semaphoreset::dontRetryInterruptedOperations() {
 	pvt->_retryinterruptedoperations=false;
 }
+
+bool semaphoreset::supported() {
+	#if defined(RUDIMENTS_HAVE_SEMGET)
+		error::clearError();
+		semget(0,0,0);
+		return (error::getErrorNumber()!=ENOSYS
+			#ifdef ENOTSUP
+				&& error::getErrorNumber()!=ENOTSUP
+			#endif
+			);
+	#elif defined(RUDIMENTS_HAVE_CREATESEMAPHORE)
+		return true;
+	#else
+		return false;
+	#endif
+}
