@@ -1,4 +1,4 @@
-// Copyright (c) 2013 David Muse
+// Copyright (c) 1999-2018 David Muse
 // See the COPYING file for more information
 
 #include <rudiments/stdio.h>
@@ -12,6 +12,16 @@ stdiofiledescriptor::stdiofiledescriptor(int32_t fd) : filedescriptor(fd) {
 	allowShortReads();
 }
 
+stdiofiledescriptor::~stdiofiledescriptor() {
+
+	// This destructor runs when a program exits, but it can also happen
+	// when a library is unloaded.  Eg. an apache module.  In cases like
+	// that, we don't want the file descriptors for 0, 1, or 2 to be closed
+	// or the server itself won't be able to read/write to them.  Setting
+	// the file descriptor to -1 here accomplishes this.
+	setFileDescriptor(-1);
+}
+
 bool stdiofiledescriptor::flush() {
 	FILE	*strm=NULL;
 	switch (fd()) {
