@@ -1,4 +1,4 @@
-// Copyright (c) 2012 David Muse
+// Copyright (c) 1999-2018 David Muse
 // See the COPYING file for more information
 
 #include <rudiments/private/config.h>
@@ -10,6 +10,12 @@
 #include <rudiments/stringbuffer.h>
 #include <rudiments/stdio.h>
 
+#ifndef RUDIMENTS_HAVE_GETHOSTNAME
+	#include <rudiments/linkedlist.h>
+	#include <rudiments/directory.h>
+	#include <rudiments/file.h>
+#endif
+
 #include <rudiments/private/winsock.h>
 
 #ifdef RUDIMENTS_HAVE_SYS_TYPES_H
@@ -78,6 +84,47 @@
 #endif
 
 #include <stdio.h>
+#ifdef RUDIMENTS_HAVE_SYS_MMAN_H
+	#include <sys/mman.h>
+#endif
+
+// if SSIZE_MAX is undefined, choose a good safe value
+// that should even work on 16-bit systems
+#ifndef SSIZE_MAX
+        #define SSIZE_MAX 16383
+#endif
+
+#ifdef RUDIMENTS_HAVE_MPROTECT_CADDR_T
+	#define MPROTECT_ADDRCAST caddr_t
+#else
+	#define MPROTECT_ADDRCAST void *
+#endif
+
+#ifdef RUDIMENTS_HAVE_MINCORE_CADDR_T
+	#define MINCORE_ADDRCAST caddr_t
+#else
+	#define MINCORE_ADDRCAST void *
+#endif
+
+
+#ifdef RUDIMENTS_HAVE_MLOCK_CADDR_T
+	#define MLOCK_ADDRCAST caddr_t
+#else
+	#define MLOCK_ADDRCAST void *
+#endif
+
+#ifdef RUDIMENTS_HAVE_MUNLOCK_CADDR_T
+	#define MUNLOCK_ADDRCAST caddr_t
+#else
+	#define MUNLOCK_ADDRCAST void *
+#endif
+
+#ifdef RUDIMENTS_HAVE_MADVISE_CADDR_T
+	#define MADVISE_ADDRCAST caddr_t
+#else
+	#define MADVISE_ADDRCAST void *
+#endif
+
 
 char *sys::getOperatingSystemName() {
 	#if defined(RUDIMENTS_HAVE_UNAME)
@@ -192,6 +239,10 @@ char *sys::getOperatingSystemArchitecture() {
 	#endif
 }
 
+const char *sys::getRudimentsVersion() {
+	return RUDIMENTS_VERSION;
+}
+
 char *sys::getHostName() {
 	#if defined(RUDIMENTS_HAVE_GETHOSTNAME)
 
@@ -232,9 +283,50 @@ char *sys::getHostName() {
 
 		// for any other error, return null
 		return NULL;
+
 	#else
-		RUDIMENTS_SET_ENOSYS
-		return NULL;
+
+		// open /etc
+		directory	d;
+		if (!d.open("/etc")) {
+			return NULL;
+		}
+
+		// get various file names into a list
+		linkedlist< char * >	nodenames;
+		for (;;) {
+			char	*f=d.read();
+			if (!charstring::compare(f,"hostname") ||
+				!charstring::compare(f,"nodename") ||
+				!charstring::compare(f,"nodename.")) {
+				nodenames.append(f);
+			} else if (!f) {
+				break;
+			} else {
+				delete[] f;
+			}
+		}
+
+		// bail if there weren't any
+		if (!nodenames.getLength()) {
+			return NULL;
+		}
+
+		// sort the list...
+		// fortunately we care about hostname, nodename, then nodename.*
+		// in that order, which also happens to be alphabetical order
+		nodenames.heapSort();
+
+		// get the contents of the first file
+		stringbuffer	fname;
+		fname.append("/etc/")->append(nodenames.getFirst()->getValue());
+		char	*hostname=file::getContents(fname.getString());
+
+		// clean up the list
+		nodenames.clearAndArrayDelete();
+
+		// return the hostname
+		return (!charstring::isNullOrEmpty(hostname))?hostname:NULL;
 	#endif
 }
 
@@ -443,7 +535,7 @@ bool sys::halt() {
 		int32_t	cmd;
 		#if defined(LINUX_REBOOT_CMD_HALT)
 			cmd=LINUX_REBOOT_CMD_HALT;
-		#elif defined(RB_HALT)
+		#elif defined(RB_HALT) || defined(RUDIMENTS_HAVE_RB_HALT)
 			cmd=RB_HALT;
 		#elif defined(RB_HALT_SYSTEM)
 			cmd=RB_HALT_SYSTEM;
@@ -455,7 +547,7 @@ bool sys::halt() {
 		return (::reboot(cmd)!=-1);
 	#elif defined(RUDIMENTS_HAVE_REBOOT_2)
 		int32_t	cmd;
-		#if defined(RB_HALT)
+		#if defined(RB_HALT) || defined(RUDIMENTS_HAVE_RB_HALT)
 			cmd=RB_HALT;
 		#else
 			#error no RB_HALT or anything like it
@@ -499,7 +591,7 @@ bool sys::shutDown() {
 			cmd=LINUX_REBOOT_CMD_POWER_OFF;
 		#elif defined(RB_POWEROFF)
 			cmd=RB_POWEROFF;
-		#elif defined(RB_HALT)
+		#elif defined(RB_HALT) || defined(RUDIMENTS_HAVE_RB_HALT)
 			// OS X doesn't appear to have a power-down option,
 			// just use halt instead
 			cmd=RB_HALT;
@@ -519,7 +611,7 @@ bool sys::shutDown() {
 		int32_t	cmd;
 		#if defined(RB_POWERDOWN)
 			cmd=RB_POWERDOWN;
-		#elif defined(RB_HALT)
+		#elif defined(RB_HALT) || defined(RUDIMENTS_HAVE_RB_HALT)
 			// Solaris doesn't appear to have a power-down option,
 			// just use halt instead
 			cmd=RB_HALT;
@@ -546,6 +638,7 @@ bool sys::shutDown() {
 		RUDIMENTS_SET_ENOSYS
 		return false;
 	#endif
+
 }
 
 bool sys::reboot() {
@@ -553,7 +646,8 @@ bool sys::reboot() {
 		int32_t	cmd;
 		#if defined(LINUX_REBOOT_CMD_RESTART)
 			cmd=LINUX_REBOOT_CMD_RESTART;
-		#elif defined(RB_AUTOBOOT)
+		#elif defined(RB_AUTOBOOT) || \
+			defined(RUDIMENTS_HAVE_RB_AUTOBOOT)
 			cmd=RB_AUTOBOOT;
 		#elif defined(RBT_REBOOT)
 			cmd=RBT_REBOOT;
@@ -563,7 +657,8 @@ bool sys::reboot() {
 		return (::reboot(cmd)!=-1);
 	#elif defined(RUDIMENTS_HAVE_REBOOT_2)
 		int32_t	cmd;
-		#if defined(RB_AUTOBOOT)
+		#if defined(RB_AUTOBOOT) || \
+			defined(RUDIMENTS_HAVE_RB_AUTOBOOT)
 			cmd=RB_AUTOBOOT;
 		#else
 			#error no RB_AUTOBOOT or anything like it
@@ -748,8 +843,7 @@ int64_t sys::getMaxLineLength() {
 	#if defined(_SC_LINE_MAX)
 		return sysConf(_SC_LINE_MAX);
 	#else
-		RUDIMENTS_SET_ENOSYS
-		return -1;
+		return SSIZE_MAX;
 	#endif
 }
 
@@ -934,7 +1028,8 @@ int64_t sys::getMaxTimersPerProcess() {
 
 int64_t sys::getSuggestedGroupEntryBufferSize() {
 	#if defined(_SC_GETGR_R_SIZE_MAX)
-		return sysConf(_SC_GETGR_R_SIZE_MAX);
+		int64_t	size=sysConf(_SC_GETGR_R_SIZE_MAX);
+		return (size>0)?size:1024;
 	#else
 		return 1024;
 	#endif
@@ -942,7 +1037,8 @@ int64_t sys::getSuggestedGroupEntryBufferSize() {
 
 int64_t sys::getSuggestedPasswordEntryBufferSize() {
 	#if defined(_SC_GETPW_R_SIZE_MAX)
-		return sysConf(_SC_GETPW_R_SIZE_MAX);
+		int64_t	size=sysConf(_SC_GETPW_R_SIZE_MAX);
+		return (size>0)?size:1024;
 	#else
 		return 1024;
 	#endif
@@ -1010,7 +1106,7 @@ int64_t sys::getCpuSetSize() {
 	#elif defined(_WIN32)
 
 		// 64 for 64-bit Windows 7/Server 2008 R2,
-		// 0 for other platforms...
+		// 1 for other platforms...
 
         	#if _WIN32_WINNT>=0x0601
 			SYSTEM_INFO	sysinfo;
@@ -1018,9 +1114,9 @@ int64_t sys::getCpuSetSize() {
 			return (sysinfo.wProcessorArchitecture==
 					PROCESSOR_ARCHITECTURE_AMD64 ||
 				sysinfo.wProcessorArchitecture==
-					PROCESSOR_ARCHITECTURE_IA64)?64:0;
+					PROCESSOR_ARCHITECTURE_IA64)?64:1;
 		#else
-			return 0;
+			return 1;
 		#endif
 	#else
 		RUDIMENTS_SET_ENOSYS
@@ -1122,3 +1218,208 @@ char sys::getDirectorySeparator() {
 		return '/';
 	#endif
 }
+
+bool sys::setProtection(unsigned char *ptr, size_t size, int32_t protection) {
+	#ifdef RUDIMENTS_HAVE_MPROTECT
+		int32_t	result;
+		error::clearError();
+		do {
+			result=mprotect(
+				reinterpret_cast<MPROTECT_ADDRCAST>(ptr),
+				size,protection);
+		} while (result==-1 && error::getErrorNumber()==EINTR);
+		return !result;
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return false;
+	#endif
+}
+
+bool sys::sequentialAccess(unsigned char *ptr, size_t size) {
+	#if defined(RUDIMENTS_HAVE_MADVISE) && defined(MADV_SEQUENTIAL)
+		return mAdvise(ptr,size,MADV_SEQUENTIAL);
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return true;
+	#endif
+}
+
+bool sys::randomAccess(unsigned char *ptr, size_t size) {
+	#if defined(RUDIMENTS_HAVE_MADVISE) && defined(MADV_RANDOM)
+		return mAdvise(ptr,size,MADV_RANDOM);
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return true;
+	#endif
+}
+
+bool sys::willNeed(unsigned char *ptr, size_t size) {
+	#if defined(RUDIMENTS_HAVE_MADVISE) && defined(MADV_WILLNEED)
+		return mAdvise(ptr,size,MADV_WILLNEED);
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return true;
+	#endif
+}
+
+bool sys::wontNeed(unsigned char *ptr, size_t size) {
+	#if defined(RUDIMENTS_HAVE_MADVISE) && defined(MADV_DONTNEED)
+		return mAdvise(ptr,size,MADV_DONTNEED);
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return true;
+	#endif
+}
+
+bool sys::normalAccess(unsigned char *ptr, size_t size) {
+	#if defined(RUDIMENTS_HAVE_MADVISE) && defined(MADV_NORMAL)
+		return mAdvise(ptr,size,MADV_NORMAL);
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return true;
+	#endif
+}
+
+bool sys::lock(unsigned char *ptr, size_t size) {
+	#ifdef RUDIMENTS_HAVE_MLOCK
+		int32_t	result;
+		error::clearError();
+		do {
+			result=mlock(
+				reinterpret_cast<MLOCK_ADDRCAST>(ptr),
+				size);
+		} while (result==-1 && error::getErrorNumber()==EINTR);
+		return !result;
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return false;
+	#endif
+}
+
+bool sys::unlock(unsigned char *ptr, size_t size) {
+	#ifdef RUDIMENTS_HAVE_MUNLOCK
+		int32_t	result;
+		error::clearError();
+		do {
+			result=munlock(
+				reinterpret_cast<MUNLOCK_ADDRCAST>(ptr),
+				size);
+		} while (result==-1 && error::getErrorNumber()==EINTR);
+		return !result;
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return false;
+	#endif
+}
+
+bool sys::notPagedOut(unsigned char *ptr, size_t size) {
+
+	#ifdef RUDIMENTS_HAVE_MINCORE
+		// create an array of char's, 1 for each page
+		int32_t		pagesize=sys::getAllocationGranularity();
+		int32_t		tmpsize=(size+pagesize-1)/pagesize;
+		#ifdef RUDIMENTS_HAVE_MINCORE_CHAR
+			char		*tmp=new char[tmpsize];
+		#else
+			unsigned char	*tmp=new unsigned char[tmpsize];
+		#endif
+
+		// call mincore to fill the array
+		int32_t		result;
+		error::clearError();
+		do {
+			result=mincore(
+				reinterpret_cast<MINCORE_ADDRCAST>(ptr),
+				size,tmp);
+		} while (result==-1 && error::getErrorNumber()==EINTR);
+		if (result) {
+			delete[] tmp;
+			return false;
+		}
+
+		// look through the array, if any of the
+		// pages aren't in memory, return false
+		for (int32_t i=0; i<tmpsize; i++) {
+			if (tmp[i]) {
+				delete[] tmp;
+				return false;
+			}
+		}
+		delete[] tmp;
+		return true;
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return false;
+	#endif
+}
+
+bool sys::disablePaging() {
+	#if defined(MCL_CURRENT) && defined(MCL_FUTURE)
+		return mLockAll(MCL_CURRENT|MCL_FUTURE);
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return false;
+	#endif
+}
+
+bool sys::disablePagingOfCurrent() {
+	#if defined(MCL_CURRENT)
+		return mLockAll(MCL_CURRENT);
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return false;
+	#endif
+}
+
+bool sys::disablePagingOfNew() {
+	#if defined(MCL_FUTURE)
+		return mLockAll(MCL_FUTURE);
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return false;
+	#endif
+}
+
+bool sys::enablePaging() {
+	#ifdef RUDIMENTS_HAVE_MUNLOCKALL
+		int32_t	result;
+		error::clearError();
+		do {
+			result=munlockall();
+		} while (result==-1 && error::getErrorNumber()==EINTR);
+		return !result;
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return false;
+	#endif
+}
+
+bool sys::mAdvise(unsigned char *start, size_t size, int32_t advice) {
+	#if defined(RUDIMENTS_HAVE_MADVISE)
+		int32_t	result;
+		error::clearError();
+		do {
+			result=madvise(
+				reinterpret_cast<MADVISE_ADDRCAST>(start),
+				size,advice);
+		} while (result==-1 && error::getErrorNumber()==EINTR);
+		return !result;
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return true;
+	#endif
+}
+
+bool sys::mLockAll(int32_t flags) {
+	#ifdef RUDIMENTS_HAVE_MLOCKALL
+		int32_t	result;
+		error::clearError();
+		do {
+			result=mlockall(flags);
+		} while (result==-1 && error::getErrorNumber()==EINTR);
+		return !result;
+	#else
+		RUDIMENTS_SET_ENOSYS
+		return false;
+	#endif
+}
