@@ -1,4 +1,4 @@
-// Copyright (c) 2016 David Muse
+// Copyright (c) 1999-2018 David Muse
 // See the COPYING file for more information.
 
 #include <rudiments/tls.h>
@@ -18,6 +18,11 @@
 	#include <rudiments/bytestring.h>
 	#include <schannel.h>
 	#define SSPI_ERROR(sstatus)	((sstatus)<0)
+
+	// MSVC 2008- doesn't define HCRYPTPROV_OR_NCRYPT_KEY_HANDLE
+	#if _MSC_VER <= 1500
+		typedef	ULONG_PTR	HCRYPTPROV_OR_NCRYPT_KEY_HANDLE;
+	#endif
 #endif
 
 #ifdef RUDIMENTS_HAVE_LIMITS_H
@@ -27,15 +32,34 @@
 // if SSIZE_MAX is undefined, choose a good safe value
 // that should even work on 16-bit systems
 #ifndef SSIZE_MAX
-	#define SSIZE_MAX 32767
+	#define SSIZE_MAX 16383
+#endif
+
+#if defined(RUDIMENTS_HAS_SSPI)
+// FIXME: move to charstring class
+static WCHAR *asciiToUnicode(const CHAR *in) {
+
+	int32_t	size=MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,in,-1,NULL,0);
+	if (!size) {
+		return NULL;
+	}
+
+	WCHAR	*out=new WCHAR[size];
+	if (!MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,in,-1,out,size)) {
+		delete[] out;
+		out=NULL;
+	}
+	return out;
+}
 #endif
 
 //#define DEBUG_TLS 1
+//#define DEBUG_CIPHERS 1
 
 threadmutex	tls::_tlsmutex;
 bool		tls::_initialized=false;
 
-bool tls::supportsTLS() {
+bool tls::supported() {
 	#if defined(RUDIMENTS_HAS_SSL)
 		return true;
 	#elif defined(RUDIMENTS_HAS_SSPI)
@@ -65,6 +89,7 @@ class tlscontextprivate {
 		bool			_isclient;
 		const char		*_version;
 		const char		*_cert;
+		const char		*_pk;
 		const char		*_pkpwd;
 		const char		*_ciphers;
 		bool			_validatepeer;
@@ -101,8 +126,13 @@ tlscontext::tlscontext() : securitycontext() {
 	pvt->_isclient=false;
 	pvt->_version=NULL;
 	pvt->_cert=NULL;
+	pvt->_pk=NULL;
 	pvt->_pkpwd=NULL;
+	#ifdef RUDIMENTS_DEFAULT_CIPHER_PROFILE_SYSTEM
+	pvt->_ciphers="PROFILE=SYSTEM";
+	#else
 	pvt->_ciphers=NULL;
+	#endif
 	pvt->_validatepeer=false;
 	pvt->_depth=9;
 	pvt->_ca=NULL;
@@ -236,6 +266,15 @@ const char *tlscontext::getCertificateChainFile() {
 	return pvt->_cert;
 }
 
+void tlscontext::setPrivateKeyFile(const char *filename) {
+	pvt->_dirty=true;
+	pvt->_pk=filename;
+}
+
+const char *tlscontext::getPrivateKeyFile() {
+	return pvt->_pk;
+}
+
 void tlscontext::setPrivateKeyPassword(const char *password) {
 	pvt->_dirty=true;
 	pvt->_pkpwd=password;
@@ -247,7 +286,12 @@ const char *tlscontext::getPrivateKeyPassword() {
 
 void tlscontext::setCiphers(const char *ciphers) {
 	pvt->_dirty=true;
+	#ifdef RUDIMENTS_DEFAULT_CIPHER_PROFILE_SYSTEM
+	pvt->_ciphers=(charstring::isNullOrEmpty(ciphers))?
+					"PROFILE=SYSTEM":ciphers;
+	#else
 	pvt->_ciphers=ciphers;
+	#endif
 }
 
 const char *tlscontext::getCiphers() {
@@ -523,7 +567,7 @@ bool tlscontext::isPeerCertValid() {
 #if defined(RUDIMENTS_HAS_SSL)
 static int passwdCallback(char *buf, int size, int rwflag, void *userdata) {
 	charstring::copy(buf,(char *)userdata,size);
-	buf[size-1]=(char)NULL;
+	buf[size-1]='\0';
 	return charstring::length(buf);
 }
 #endif 
@@ -564,18 +608,26 @@ static void getCipherAlgs(const char *ciphers, bool isclient,
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"3DES_112")) {
 			(*algids)[algindex++]=CALG_3DES_112;
+		#ifdef CALG_AES
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"AES")) {
 			(*algids)[algindex++]=CALG_AES;
+		#endif
+		#ifdef CALG_AES_128
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"AES_128")) {
 			(*algids)[algindex++]=CALG_AES_128;
+		#endif
+		#ifdef CALG_AES_192
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"AES_192")) {
 			(*algids)[algindex++]=CALG_AES_192;
+		#endif
+		#ifdef CALG_AES_256
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"AES_256")) {
 			(*algids)[algindex++]=CALG_AES_256;
+		#endif
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"AGREEDKEY_ANY")) {
 			(*algids)[algindex++]=CALG_AGREEDKEY_ANY;
@@ -612,12 +664,16 @@ static void getCipherAlgs(const char *ciphers, bool isclient,
 						c[i],"ECDSA")) {
 			(*algids)[algindex++]=CALG_ECDSA;
 		#endif
+		#ifdef CALG_ECMQV
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"ECMQV")) {
 			(*algids)[algindex++]=CALG_ECMQV;
+		#endif
+		#ifdef CALG_HASH_REPLACE_OWF
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"HASH_REPLACE_OWF")) {
 			(*algids)[algindex++]=CALG_HASH_REPLACE_OWF;
+		#endif
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"HUGHES_MD5")) {
 			(*algids)[algindex++]=CALG_HUGHES_MD5;
@@ -639,15 +695,21 @@ static void getCipherAlgs(const char *ciphers, bool isclient,
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"MD5")) {
 			(*algids)[algindex++]=CALG_MD5;
+		#ifdef CALG_NO_SIGN
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"NO_SIGN")) {
 			(*algids)[algindex++]=CALG_NO_SIGN;
+		#endif
+		#ifdef CALG_OID_INFO_CNG_ONLY
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"OID_INFO_CNG_ONLY")) {
 			(*algids)[algindex++]=CALG_OID_INFO_CNG_ONLY;
+		#endif
+		#ifdef CALG_OID_INFO_PARAMETERS
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"OID_INFO_PARAMETERS")) {
 			(*algids)[algindex++]=CALG_OID_INFO_PARAMETERS;
+		#endif
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"PCT1_MASTER")) {
 			(*algids)[algindex++]=CALG_PCT1_MASTER;
@@ -684,15 +746,21 @@ static void getCipherAlgs(const char *ciphers, bool isclient,
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"SHA1")) {
 			(*algids)[algindex++]=CALG_SHA1;
+		#ifdef CALG_SHA_256
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"SHA_256")) {
 			(*algids)[algindex++]=CALG_SHA_256;
+		#endif
+		#ifdef CALG_SHA_384
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"SHA_384")) {
 			(*algids)[algindex++]=CALG_SHA_384;
+		#endif
+		#ifdef CALG_SHA_512
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"SHA_512")) {
 			(*algids)[algindex++]=CALG_SHA_512;
+		#endif
 		} else if (!charstring::compareIgnoringCase(
 						c[i],"SKIPJACK")) {
 			(*algids)[algindex++]=CALG_SKIPJACK;
@@ -749,6 +817,34 @@ static void getCipherAlgs(const char *ciphers, bool isclient,
 }
 #endif
 
+#ifdef RUDIMENTS_HAS_SSPI
+static DWORD getLocation(const char *provider) {
+	if (!charstring::compare(provider,
+				"LOCAL_MACHINE")) {
+    		return CERT_SYSTEM_STORE_LOCAL_MACHINE;
+	} else if (!charstring::compare(provider,
+				"CURRENT_SERVICE")) {
+    		return CERT_SYSTEM_STORE_CURRENT_SERVICE;
+	} else if (!charstring::compare(provider,
+				"SERVICES")) {
+    		return CERT_SYSTEM_STORE_SERVICES;
+	} else if (!charstring::compare(provider,
+				"USERS")) {
+    		return CERT_SYSTEM_STORE_USERS;
+	} else if (!charstring::compare(provider,
+				"CURRENT_USER_GROUP_POLICY")) {
+    		return CERT_SYSTEM_STORE_CURRENT_USER_GROUP_POLICY;
+	} else if (!charstring::compare(provider,
+				"LOCAL_MACHINE_GROUP_POLICY")) {
+    		return CERT_SYSTEM_STORE_LOCAL_MACHINE_GROUP_POLICY;
+	} else if (!charstring::compare(provider,
+				"LOCAL_MACHINE_ENTERPRISE")) {
+    		return CERT_SYSTEM_STORE_LOCAL_MACHINE_ENTERPRISE;
+	}
+    	return CERT_SYSTEM_STORE_CURRENT_USER;
+}
+#endif
+
 bool tlscontext::reInit(bool isclient) {
 
 	// free/clear any old peer certificate
@@ -801,7 +897,9 @@ bool tlscontext::reInit(bool isclient) {
 		// decide on the protocol version
 		const SSL_METHOD	*method=NULL;
 		if (!charstring::compareIgnoringCase(
-					pvt->_version,"SSL2")) {
+					pvt->_version,"SSL2") ||
+			!charstring::compareIgnoringCase(
+					pvt->_version,"SSLv2")) {
 			#if defined(RUDIMENTS_HAS_SSLV2_METHOD)
 			if (isclient) {
 				method=SSLv2_client_method();
@@ -810,7 +908,9 @@ bool tlscontext::reInit(bool isclient) {
 			}
 			#endif
 		} else if (!charstring::compareIgnoringCase(
-					pvt->_version,"SSL3")) {
+					pvt->_version,"SSL3") ||
+			!charstring::compareIgnoringCase(
+					pvt->_version,"SSLv3")) {
 			#if defined(RUDIMENTS_HAS_SSLV3_METHOD)
 			if (isclient) {
 				method=SSLv3_client_method();
@@ -819,7 +919,9 @@ bool tlscontext::reInit(bool isclient) {
 			}
 			#endif
 		} else if (!charstring::compareIgnoringCase(
-					pvt->_version,"TLS1")) {
+					pvt->_version,"TLS1") ||
+			!charstring::compareIgnoringCase(
+					pvt->_version,"TLSv1")) {
 			#if defined(RUDIMENTS_HAS_TLSV1_METHOD)
 			if (isclient) {
 				method=TLSv1_client_method();
@@ -828,7 +930,9 @@ bool tlscontext::reInit(bool isclient) {
 			}
 			#endif
 		} else if (!charstring::compareIgnoringCase(
-					pvt->_version,"TLS1.1")) {
+					pvt->_version,"TLS1.1") ||
+			!charstring::compareIgnoringCase(
+					pvt->_version,"TLSv1.1")) {
 			#if defined(RUDIMENTS_HAS_TLSV1_1_METHOD)
 			if (isclient) {
 				method=TLSv1_1_client_method();
@@ -837,7 +941,9 @@ bool tlscontext::reInit(bool isclient) {
 			}
 			#endif
 		} else if (!charstring::compareIgnoringCase(
-					pvt->_version,"TLS1.2")) {
+					pvt->_version,"TLS1.2") ||
+			!charstring::compareIgnoringCase(
+					pvt->_version,"TLSv1.2")) {
 			#if defined(RUDIMENTS_HAS_TLSV1_2_METHOD)
 			if (isclient) {
 				method=TLSv1_2_client_method();
@@ -846,25 +952,41 @@ bool tlscontext::reInit(bool isclient) {
 			}
 			#endif
 		}
+
+		// NOTE: the above XXX_client/server_method calls are
+		// deprecated.  Eg. TLSv1_3_xxx_method() doesn't exist at all.
+		//
+		// OpenSSL wants you to call whichever of the methods below
+		// exist, and let both sides negotiate the highest supported
+		// version.
+		//
+		// However, this can lead to odd behavior with different
+		// versions of OpenSSL.  Eg. if you have an app that requests
+		// SSL3 on a platform that supports TLS and SSL3, and has
+		// SSLv3_xxx_method() then it will use SSL3.  But, on a
+		// platform that supports TLS and SSL3 but doesn't have
+		// SSLv3_xxx_method(), the same app will end up using the
+		// highest TLS method available, despite specifically
+		// requesting SSL3, and working as expected on the older
+		// platform.
+
 		if (!method) {
 			if (isclient) {
-				method=
 				#if defined(RUDIMENTS_HAS_TLS_METHOD)
-					TLS_client_method()
+					method=TLS_client_method()
 				#elif defined(RUDIMENTS_HAS_SSLV23_METHOD)
-					SSLv23_client_method()
+					method=SSLv23_client_method()
 				#elif defined(RUDIMENTS_HAS_SSLV2_METHOD)
-					SSLv2_client_method()
+					method=SSLv2_client_method()
 				#endif
 				;
 			} else {
-				method=
 				#if defined(RUDIMENTS_HAS_TLS_METHOD)
-					TLS_server_method()
+					method=TLS_server_method()
 				#elif defined(RUDIMENTS_HAS_SSLV23_METHOD)
-					SSLv23_server_method()
+					method=SSLv23_server_method()
 				#elif defined(RUDIMENTS_HAS_SSLV2_METHOD)
-					SSLv2_server_method()
+					method=SSLv2_server_method()
 				#endif
 				;
 			}
@@ -892,9 +1014,11 @@ bool tlscontext::reInit(bool isclient) {
 			if (SSL_CTX_use_certificate_chain_file(
 						pvt->_ctx,pvt->_cert)!=1 ||
 				SSL_CTX_use_PrivateKey_file(
-						pvt->_ctx,pvt->_cert,
+						pvt->_ctx,
+						(pvt->_pk)?pvt->_pk:pvt->_cert,
 						SSL_FILETYPE_PEM)!=1 ||
 				SSL_CTX_check_private_key(pvt->_ctx)!=1) {
+				setError(0);
 				retval=false;
 			}
 			#ifdef DEBUG_TLS
@@ -905,7 +1029,9 @@ bool tlscontext::reInit(bool isclient) {
 		// set ciphers
 		if (retval && !charstring::isNullOrEmpty(pvt->_ciphers)) {
 			if (SSL_CTX_set_cipher_list(
-					pvt->_ctx,pvt->_ciphers)!=1) {
+					pvt->_ctx,
+					pvt->_ciphers)!=1) {
+				setError(0);
 				retval=false;
 			}
 		}
@@ -925,6 +1051,7 @@ bool tlscontext::reInit(bool isclient) {
 					pvt->_ctx,
 					(ispath)?NULL:pvt->_ca,
 					(ispath)?pvt->_ca:NULL)!=1) {
+				setError(0);
 				retval=false;
 			}
 		}
@@ -946,9 +1073,11 @@ bool tlscontext::reInit(bool isclient) {
 
 		#ifdef DEBUG_TLS
 			stdoutput.printf("  version: %s\n",pvt->_version);
-			stdoutput.printf("  cert chain: %s - (%s) - (%s)\n",
-					pvt->_cert,pvt->_pkpwd,
-					(certloaded)?"valid":"invalid");
+			stdoutput.printf("  cert chain: %s\n",pvt->_cert);
+			stdoutput.printf("  private key: %s\n",
+					(pvt->_pk)?pvt->_pk:pvt->_cert);
+			stdoutput.printf("  cert/pk loaded/valid: %s\n",
+					(certloaded)?"yes":"no");
 			stdoutput.printf("  ciphers: %s\n",pvt->_ciphers);
 			stdoutput.printf("  depth: %d\n",pvt->_depth);
 			stdoutput.printf("  ca: %s\n",pvt->_ca);
@@ -1104,7 +1233,7 @@ bool tlscontext::reInit(bool isclient) {
 			}
 		}
 
-		// set certificate chain file
+		// set the certificate and private key
 		if (!charstring::isNullOrEmpty(pvt->_cert)) {
 
 			// cert specified...
@@ -1114,7 +1243,9 @@ bool tlscontext::reInit(bool isclient) {
 				credflags|=SCH_CRED_NO_DEFAULT_CREDS;
 			}
 
-			// open the specified certificate store
+			// open the specified certificate store...
+
+			// first look for a file named pvt->_cert
 			pvt->_cstore=CertOpenStore(
 					CERT_STORE_PROV_FILENAME_A,
 					X509_ASN_ENCODING|
@@ -1160,6 +1291,100 @@ bool tlscontext::reInit(bool isclient) {
 					pvt->_cctx[0]=cctx;
 					pvt->_cctxcount=1;
 				}
+
+			} else {
+
+				// clear any error set by the previous call to
+				// CertOpenStore()
+				error::clearNativeError();
+
+				// if there was no cert file with the specified
+				// name, then look for the cert in the system
+				// store...
+
+				// assume that pvt->_cert is formatted like:
+				// * location:store:subject
+				// or
+				// * store:subject
+				// 	(location presumed to be
+				//	CERT_SYSTEM_STORE_CURRENT_USER)
+				// or
+				// * subject
+				// 	(location presumed to be
+				//	CERT_SYSTEM_STORE_CURRENT_USER,
+				//	store presumed to be "MY")
+
+				// split/process pvt->_cert 
+				char		**parts=NULL;
+				uint64_t	partcount=0;
+				charstring::split(pvt->_cert,":",true,
+							&parts,&partcount);
+				DWORD	location=CERT_SYSTEM_STORE_CURRENT_USER;
+				const char	*store="MY";
+				const char	*subject="";
+				if (partcount>2) {
+					location=getLocation(parts[0]);
+					store=parts[1];
+					subject=parts[2];
+				} else if (partcount==2) {
+					store=parts[0];
+					subject=parts[1];
+				} else if (partcount==1) {
+					subject=parts[0];
+				}
+
+				// first look for the specified store
+				pvt->_cstore=CertOpenStore(
+						CERT_STORE_PROV_SYSTEM_A,
+						0,
+						NULL,
+						location,
+						store);
+
+				if (pvt->_cstore) {
+
+					// get the cert who's subject
+					// includes the specified string
+					WCHAR	*subjectw=
+					asciiToUnicode(subject);
+					PCCERT_CONTEXT	cctx=
+					CertFindCertificateInStore(
+							pvt->_cstore,
+							X509_ASN_ENCODING|
+							PKCS_7_ASN_ENCODING,
+							0,
+							CERT_FIND_SUBJECT_STR,
+							subjectw,
+							NULL);
+					delete[] subjectw;
+
+					// FIXME: support private key password
+
+					if (cctx) {
+
+						// verify that the cert
+						// contained a private key
+					HCRYPTPROV_OR_NCRYPT_KEY_HANDLE	kh=NULL;
+					CryptAcquireCertificatePrivateKey(
+								cctx,
+								0,
+								NULL,
+								&kh,
+								NULL,
+								NULL);
+
+						pvt->_cctx=
+						new PCCERT_CONTEXT[1];
+						pvt->_cctx[0]=cctx;
+						pvt->_cctxcount=1;
+					}
+				}
+
+				// clean up 
+				for (uint64_t i=0; i<partcount; i++) {
+					delete[] parts[i];
+				}
+				delete[] parts;
 			}
 
 			// If CertOpenStoreFailed, CertFindCertificateInStore
@@ -1194,6 +1419,54 @@ bool tlscontext::reInit(bool isclient) {
 					NULL,
 					0,
 					pvt->_ca);
+
+			if (!pvt->_castore) {
+
+				// clear any error set by the previous call to
+				// CertOpenStore()
+				error::clearNativeError();
+
+				// if there was no ca file with the specified
+				// name, then look for the ca in the system
+				// store...
+
+				// assume that pvt->_ca is formatted like:
+				// * location:store
+				// or
+				// * store:subject
+				// 	(location presumed to be
+				//	CERT_SYSTEM_STORE_CURRENT_USER)
+
+				// split/process pvt->_ca 
+				char		**parts=NULL;
+				uint64_t	partcount=0;
+				charstring::split(pvt->_ca,":",true,
+							&parts,&partcount);
+				DWORD	location=CERT_SYSTEM_STORE_CURRENT_USER;
+				const char	*store="MY";
+				if (partcount>1) {
+					location=getLocation(parts[0]);
+					store=parts[1];
+				} else if (partcount==1) {
+					store=parts[0];
+				}
+
+				// open the specified certificate
+				// authority store
+				pvt->_castore=CertOpenStore(
+						CERT_STORE_PROV_SYSTEM_A,
+						0,
+						NULL,
+						location,
+						store);
+
+				// clean up 
+				for (uint64_t i=0; i<partcount; i++) {
+					delete[] parts[i];
+				}
+				delete[] parts;
+			}
+
 			if (!pvt->_castore) {
 				setNativeError();
 				retval=false;
@@ -1383,18 +1656,24 @@ bool tlscontext::loadPeerCert() {
 	#endif
 }
 
-ssize_t tlscontext::read(void *buf, ssize_t count) {
+ssize_t tlscontext::read(void *buf, ssize_t size) {
 	clearError();
 	#if defined(RUDIMENTS_HAS_SSL)
 		if (!pvt->_ssl) {
 			return RESULT_ERROR;
 		}
-		int	ret=SSL_read(pvt->_ssl,buf,count);
+		int	ret=SSL_read(pvt->_ssl,
+					#ifdef RUDIMENTS_SSL_VOID_PTR
+					buf,
+					#else
+					(char *)buf,
+					#endif
+					size);
 		setError(ret);
 		return ret;
 	#elif defined(RUDIMENTS_HAS_SSPI)
-		ssize_t	ret=pvt->_gctx.read(buf,count);
-		if (ret!=count) {
+		ssize_t	ret=pvt->_gctx.read(buf,size);
+		if (ret!=size) {
 			setError(ret);
 		}
 		return ret;
@@ -1404,18 +1683,24 @@ ssize_t tlscontext::read(void *buf, ssize_t count) {
 	#endif
 }
 
-ssize_t tlscontext::write(const void *buf, ssize_t count) {
+ssize_t tlscontext::write(const void *buf, ssize_t size) {
 	clearError();
 	#if defined(RUDIMENTS_HAS_SSL)
 		if (!pvt->_ssl) {
 			return RESULT_ERROR;
 		}
-		int	ret=SSL_write(pvt->_ssl,buf,count);
+		int	ret=SSL_write(pvt->_ssl,
+					#ifdef RUDIMENTS_SSL_VOID_PTR
+					buf,
+					#else
+					(const char *)buf,
+					#endif
+					size);
 		setError(ret);
 		return ret;
 	#elif defined(RUDIMENTS_HAS_SSPI)
-		ssize_t	ret=pvt->_gctx.write(buf,count);
-		if (ret!=count) {
+		ssize_t	ret=pvt->_gctx.write(buf,size);
+		if (ret!=size) {
 			setError(ret);
 		}
 		return ret;
@@ -1486,26 +1771,99 @@ void tlscontext::clearError() {
 
 void tlscontext::setError(int32_t ret) {
 	#if defined(RUDIMENTS_HAS_SSL)
-		if (!pvt->_ssl || ret==1) {
+
+		if (ret==1) {
 			return;
 		}
+
 		pvt->_errorstr.clear();
+
+		// try to get the error from ERR...
+		// (there may be multiple errors)
 		bool	first=true;
 		for (;;) {
+
+			unsigned long	err=ERR_get_error();
+			if (!err) {
+				break;
+			}
+
 			if (first) {
 				first=false;
 			} else {
 				pvt->_errorstr.append('\n');
 			}
-			unsigned long	err=ERR_get_error();
-			if (!err) {
-				break;
-			}
+
 			pvt->_error=err;
 			pvt->_errorstr.append(
 				ERR_error_string(pvt->_error,NULL));
+
+#ifdef DEBUG_CIPHERS
+			// if the error was "no shared cipher" then include
+			// the protocol and ciphers that were offered...
+			if (pvt->_ssl && charstring::contains(
+						pvt->_errorstr.getString(),
+						"no shared cipher")) {
+
+				// get the protocol from the session and
+				// include it...
+				pvt->_errorstr.append("\nprotocol: ");
+				SSL_SESSION	*sess=
+						SSL_get_session(pvt->_ssl);
+				if (sess) {
+					pvt->_errorstr.writeFormatted(
+					"0x%04x",
+					SSL_SESSION_get_protocol_version(sess));
+				} else {
+					pvt->_errorstr.append("unknown");
+				}
+				pvt->_errorstr.append('\n');
+
+				// get the offered ciphers and include them...
+				pvt->_errorstr.append("ciphers offered:\n");
+				STACK_OF(SSL_CIPHER)	*cs=
+					(STACK_OF(SSL_CIPHER) *)
+					SSL_get_client_ciphers(pvt->_ssl);
+				if (cs) {
+					for (int i=0;
+						i<sk_SSL_CIPHER_num(cs);
+						i++) {
+						if (i) {
+							pvt->_errorstr.
+								append(':');
+						}
+						pvt->_errorstr.append(
+							SSL_CIPHER_get_name(
+							sk_SSL_CIPHER_value(
+							cs,i)));
+					}
+				}
+				pvt->_errorstr.append('\n');
+
+				// get the available ciphers and include them...
+				pvt->_errorstr.append("ciphers available:\n");
+				cs=(STACK_OF(SSL_CIPHER) *)
+					SSL_CTX_get_ciphers(pvt->_ctx);
+				if (cs) {
+					for (int i=0;
+						i<sk_SSL_CIPHER_num(cs);
+						i++) {
+						if (i) {
+							pvt->_errorstr.
+								append(':');
+						}
+						pvt->_errorstr.append(
+							SSL_CIPHER_get_name(
+							sk_SSL_CIPHER_value(
+							cs,i)));
+					}
+				}
+			}
+#endif
 		}
-		if (!pvt->_errorstr.getStringLength()) {
+
+		// if that failed then try to get the error from SSL...
+		if (!pvt->_errorstr.getStringLength() && pvt->_ssl) {
 			const char	*str="";
 			switch (SSL_get_error(pvt->_ssl,ret)) {
 				case SSL_ERROR_NONE:
@@ -1685,11 +2043,11 @@ const unsigned char *tlscertificate::getPublicKey() {
 	return pvt->_pk;
 }
 
-uint64_t tlscertificate::getPublicKeyByteLength() {
+uint64_t tlscertificate::getPublicKeyByteSize() {
 	return pvt->_pklen;
 }
 
-uint64_t tlscertificate::getPublicKeyBitLength() {
+uint64_t tlscertificate::getPublicKeyBitSize() {
 	return pvt->_pkbits;
 }
 
@@ -1791,7 +2149,13 @@ void tlscertificate::setCertificate(void *cert) {
 
 		// get signature algorithm
 		pvt->_sigalg=new char[256];
-		OBJ_obj2txt(pvt->_sigalg,256,c->sig_alg->algorithm,0);
+		OBJ_obj2txt(pvt->_sigalg,256,
+			#ifdef RUDIMENTS_HAS_X509_GET_SIGNATURE_NID
+				OBJ_nid2obj(X509_get_signature_nid(c)),
+			#else
+				c->sig_alg->algorithm,
+			#endif
+				0);
 		// FIXME: signature algorithm parameters
 
 		// get the issuer
@@ -1811,14 +2175,25 @@ void tlscertificate::setCertificate(void *cert) {
 		// get the public key algorithm
 		EVP_PKEY	*pubkey=X509_get_pubkey(c);
 		pvt->_pkalg=charstring::duplicate(
-				(pubkey)?OBJ_nid2ln(pubkey->type):NULL);
+					(pubkey)?
+					OBJ_nid2ln(
+					#ifdef RUDIMENTS_HAS_EVP_PKEY_BASE_ID
+						EVP_PKEY_base_id(pubkey)
+					#else
+						pubkey->type
+					#endif
+					):NULL);
 		// FIXME: public key algorithm parameters
 
 		// get the public key
 		pvt->_pklen=EVP_PKEY_size(pubkey);
 		pvt->_pk=(unsigned char *)bytestring::duplicate(
-						pubkey->pkey.ptr,
-						pvt->_pklen);
+					#ifdef RUDIMENTS_HAS_EVP_PKEY_GET0
+						EVP_PKEY_get0(pubkey)
+					#else
+						pubkey->pkey.ptr
+					#endif
+					,pvt->_pklen);
 		pvt->_pkbits=EVP_PKEY_bits(pubkey);
 		EVP_PKEY_free(pubkey);
 
@@ -1839,21 +2214,27 @@ void tlscertificate::setCertificate(void *cert) {
 				// FIXME: get other extensions
 				if (cur->type==GEN_DNS) {
 					ASN1_IA5STRING	*dnsia5;
-					#ifdef RUDIMENTS_HAS_SSL_DNSNAME
-						dnsia5=cur->d.dNSName;
-					#else
-						dnsia5=cur->d.ia5;
-					#endif
-					char	*dnsname=(char *)
+				#ifdef RUDIMENTS_HAS_SSL_DNSNAME
+					dnsia5=cur->d.dNSName;
+				#else
+					dnsia5=cur->d.ia5;
+				#endif
+				#ifdef RUDIMENTS_HAS_ASN1_STRING_GET0_DATA
+					const unsigned char	*dnsname=
+						ASN1_STRING_get0_data(dnsia5);
+				#else
+					char *dnsname=(char *)
 						ASN1_STRING_data(dnsia5);
+				#endif
 					size_t	asn1len=
 						ASN1_STRING_length(dnsia5);
 					size_t	strlen=
-						charstring::length(dnsname);
+						charstring::length(
+							(const char *)dnsname);
 					if (asn1len==strlen) {
 						pvt->_san.append(
 							charstring::duplicate(
-								dnsname));
+							(const char *)dnsname));
 					}
 				}
 			}
@@ -1902,13 +2283,25 @@ void tlscertificate::setCertificate(void *cert) {
 		uint64_t	t=
 		(((((uint64_t)c->NotBefore.dwHighDateTime)<<32)|
 				c->NotBefore.dwLowDateTime)/10)-
-				11644473600000000ULL;
+				// Visual C++ 2002- doesn't like the ULL suffix
+				#if defined(_MSC_VER) && (_MSC_VER <= 1300)
+					11644473600000000
+				#else
+					11644473600000000ULL
+				#endif
+				;
 		pvt->_validfrom.initialize(t/1000000,t%1000000);
 
 		// get valid-to
 		t=(((((uint64_t)c->NotAfter.dwHighDateTime)<<32)|
 				c->NotAfter.dwLowDateTime)/10)-
-				11644473600000000ULL;
+				// Visual C++ 2002- doesn't like the ULL suffix
+				#if defined(_MSC_VER) && (_MSC_VER <= 1300)
+					11644473600000000
+				#else
+					11644473600000000ULL
+				#endif
+				;
 		pvt->_validto.initialize(t/1000000,t%1000000);
 
 		// get the subject
