@@ -1,7 +1,6 @@
-// Copyright (c) 2001  David Muse
+// Copyright (c) 1999-2018 David Muse
 // See the file COPYING for more information
 
-#include <rudiments/signalclasses.h>
 #include <rudiments/process.h>
 #include <rudiments/charstring.h>
 #include <rudiments/error.h>
@@ -31,7 +30,7 @@ int main(int argc, const char **argv) {
 
 	pid_t	pid=charstring::toInteger(argv[2]);
 
-	if (!signalmanager::sendSignal(pid,sig)) {
+	if (!process::sendSignal(pid,sig)) {
 		stdoutput.printf("kill failed: %s\n",error::getErrorString());
 		process::exit(1);
 	}
